#ifndef CPPOBJECT_H
#define CPPOBJECT_H

#ifdef __cplusplus
#include <atl/object.h>
#include <iostream>
#include <sstream>
#include <string.h>
#include <utils/utils.hpp>

namespace cpptesk {
	namespace atl {

//
//      
//

///    (  ).      .
///         .   
///    .
template <typename T>
struct _Storage {
	///    .  ,   ,      storgage_t.
	///        .
	typedef T* storage_t;

	///  
	static const size_t size = sizeof(storage_t);

	///     .
	static T& ref(storage_t& r) { return *r; }
	///      .
	static const T& ref(const storage_t& r) { return *r; }
	///     .
	static T* ptr(storage_t& r) { return r; }
	///      .
	static const T* ptr(const storage_t& r) { return r; }

	///   .   _Storage   , (T*)0.
	static storage_t empty() { return 0; }
	static void makeEmpty(storage_t * place) { *place = 0; }
	/// ,   .
	///
	static bool isEmpty(const storage_t& val) { return val == 0; }
	static storage_t default_ref() { return new T(); }
	static void destroy(storage_t& val) { delete val;  makeEmpty(&val); }
	static storage_t clone(const storage_t& src) { return new T(ref(src)); }

	//   
	static storage_t wrap_const(T const& src) { return new T(src); }
	//    .   
	//       destroy
	static storage_t wrap(T* src) { return src; }
	//   
	static storage_t fromNothing() { return new T(); }
	//  .
	static void assign(storage_t& dst, const storage_t& src) {
		if (!isEmpty(dst) && !isEmpty(src)) ref(dst) = ref(src);
		else if (isEmpty(src)) destroy(dst);
		else if (isEmpty(dst)) dst = clone(src);
	}
};

//   .     .
//       .
//      ++     .
template<class U>
struct _InPlace {
	typedef U storage_t;

	static U& ref(U& r) { return r; }
	static U const& ref(U const& r) { return r; }
	static U* ptr(U& r) { return &r; }
	static const U* ptr(const U& r) { return &r; }

	static U empty() { return U(); }
	static void makeEmpty(storage_t * place) {
		new(place) U();
	}
	static bool isEmpty(const U& val) { cpptesk::utils::_ignore(val); return false; }
	static U default_ref() { return U(); }
	static void destroy(U& val) { val.~U();  makeEmpty(&val); }
	static U clone(const U& src) { return U(src); }

	//   
	static U wrap_const(const U& src) { return U(src); }
	//    .   
	//       destroy
	static U wrap(U* src) { U result(*src); delete src; return result; }
	//   
	static U fromNothing() { return U(); }
	//  .
	static void assign(storage_t& dst, const storage_t& src) {
		dst = src;
	}
};

template<>
struct _Storage<bool> : public _InPlace<bool> {};

template<>
struct _Storage<char> : public _InPlace<char> {};

template<>
struct _Storage<short> : public _InPlace<short> {};

template<>
struct _Storage<int> : public _InPlace<int> {};

template<>
struct _Storage<long> : public _InPlace<long> {};

template<>
struct _Storage<unsigned char> : public _InPlace<unsigned char> {};

template<>
struct _Storage<unsigned short> : public _InPlace<unsigned short> {};

template<>
struct _Storage<unsigned int> : public _InPlace<unsigned int> {};

template<>
struct _Storage<unsigned long> : public _InPlace<unsigned long> {};


template<>
struct _Storage<float> : public _InPlace<float> {};

template<>
struct _Storage<double> : public _InPlace<double> {};

template<>
struct _Storage<long double> : public _InPlace<long double> {};

template<>
struct _Storage<std::string> : public _InPlace<std::string> {};

template<typename T>
struct _Storage<const T *> : public _InPlace<const T *> {};

//   .     .
//        .
//      ,    .
//     .
template <typename T>
struct SharedPtr {
	T* ptr;
	int count;
	SharedPtr(T* ref) : ptr(ref), count(1) {}

	SharedPtr* inc() { count ++; return this; }
	SharedPtr* dec() { if (count > 1) { -- count; return this; } else { delete this; return 0; } }
private :
	SharedPtr(const SharedPtr& other) {}
	SharedPtr& operator =  (const SharedPtr& other) {}
};

template<class U>
struct _Shared {
	typedef SharedPtr<U> _SharedPtr;
	typedef _SharedPtr* storage_t;

	static U*& ref(storage_t& r) { return r->ptr; }
	static U* const & ref(storage_t const& r) { return r->ptr; }
	static U** ptr(storage_t& r) { return &(r->ptr); }
	static const U** ptr(const storage_t& r) { return &(r->ptr); }

	static storage_t empty() { return 0; }
	static storage_t* makeEmpty(storage_t* place) { *place = 0; return place; }
	static bool isEmpty(const storage_t& val) { return val == 0; }
	static storage_t default_ref() { return new _SharedPtr(0); }
	static void destroy(storage_t& val) { val->dec(); makeEmpty(&val); }
	static storage_t clone(const storage_t& src) { return src->inc(); }

	//   
	static storage_t wrap_const(U* const& src) { return new _SharedPtr(src); }
	//    .   
	//       destroy
	static storage_t wrap(U** src) { return new _SharedPtr(*src); }
	//   
	static storage_t fromNothing() { return new _SharedPtr(); }
	//  .
	static void assign(storage_t& dst, const storage_t& src) {
		if (!isEmpty(dst)) { dst->dec();}
		if (!isEmpty(src)) { dst = src->inc(); }
		else { dst = empty(); }
	}
};

template<typename T>
struct _Storage<T*> : public _Shared<T> {};

//
//    ,   ++     
//

template <typename T>
struct CppObject {
	typedef _Storage<T> Storage;
	typedef typename Storage::storage_t storage_t;
	typedef T value_t;
public:
	static const Type type;
	storage_t value;

	T& ref() { return Storage::ref(value); }
	const T& ref() const { return Storage::ref(value); }
	T* ptr() { return Storage::ptr(value); }

	CppObject* r() { return reinterpret_cast<CppObject*>(::r(this)); }

	bool operator == (const CppObject& other) {
		if (Storage::isEmpty(value) && Storage::isEmpty(other.value)) {
			return true;
		} else if (!Storage::isEmpty(value) && !Storage::isEmpty(other.value)) {
			return (ref() == other.ref());
		} else {
			return false;
		}
	}

	int compare(const CppObject& right) {
		if (operator ==(right)) return 0;
		else return -1;
	}
	bool checkInvariant() { return true; }
	void enumerate(EnumerationCallbackFuncType f, void* param) {}

	CppObject* clone() {
		return (CppObject*)::clone_Object((Object*)(this->r()));
	}

	//  ,   ,        
	//    
	static CppObject * create(T* value) {
		CppObject<T>* obj = (CppObject<T>*)create_Object(&type);
//		Storage::makeEmpty(&(obj->value));
		obj->value = Storage::wrap(value);
		return obj;
	}
	static CppObject * create(T const& value) {
		CppObject<T>* obj = (CppObject<T>*)create_Object(&type);
//		Storage::makeEmpty(&(obj->value));
		obj->value = Storage::wrap_const(value);
		return obj;
	}
	static CppObject * create() {
		CppObject<T>* obj = (CppObject<T>*)create_Object(&type);
//		Storage::makeEmpty(&(obj->value));
		obj->value = Storage::default_ref();
		return obj;
	}
	static CppObject * empty() {
		CppObject<T>* obj = (CppObject<T>*)create_Object(&type);
//		Storage::makeEmpty(&(obj->value));
//		obj->value = Storage::empty();
		return obj;
	}
};

template <typename T>
std::ostream& operator << (std::ostream& out, const CppObject<T>& obj) {
	return out << obj.ref();
}

template <typename T> void init_CppObject(void* ref, va_list* arg_list) {
    cpptesk::utils::_ignore(arg_list);
	CppObject<T>* obj = reinterpret_cast<CppObject<T>*>(ref);
	CppObject<T>::Storage::makeEmpty(&(obj->value));
//	obj->value = CppObject<T>::Storage::empty();
}

template <typename T> void copy_CppObject(void* src, void* dst) {
	CppObject<T>* src_obj = reinterpret_cast<CppObject<T>*>(src);
	CppObject<T>* dst_obj = reinterpret_cast<CppObject<T>*>(dst);
	CppObject<T>::Storage::makeEmpty(&(dst_obj->value));
	CppObject<T>::Storage::assign(dst_obj->value, src_obj->value);
}

//template <typename T> int compare(const T& left, const T& right) {
//	return (left == right)? 0 : 1;
//}
//
template <typename T> int compare_CppObject(void * left, void * right) {
	CppObject<T>* lobj = reinterpret_cast<CppObject<T>*>(left);
	CppObject<T>* robj = reinterpret_cast<CppObject<T>*>(right);
	return lobj->compare(*robj);
}

template <typename T> bool checkInvariant_CppObject(void * arg) {
	CppObject<T>* obj = reinterpret_cast<CppObject<T>*>(arg);
	return obj->checkInvariant();
}

template <typename T> String* toString_CppObject(void * arg) {
	CppObject<T>* obj = reinterpret_cast<CppObject<T>*>(arg);
	std::ostringstream out;
	out << TYPE(arg)->name << '(';
	T& ref = obj->ref();
	out << ref;
	out << ')';
	return create_String(out.str().c_str());
}

template <typename T> String* toXML_CppObject(void* arg) {
	CppObject<T>* obj = reinterpret_cast<CppObject<T>*>(arg);
	std::ostringstream out;
	out << "<object kind=\"spec\" type=\"" << TYPE(arg)->name << "\" text=\"" << obj->ref() <<"\"/>";
	return create_String(out.str().c_str());
}

template <typename T> void enumerate_CppObject(void * arg, EnumerationCallbackFuncType f, void * param) {
	CppObject<T>* obj = reinterpret_cast<CppObject<T>*>(arg);
	return obj->enumerate(f, param);
}

template <typename T> void destroy_CppObject(void* ptr) {
	CppObject<T>* src_obj = reinterpret_cast<CppObject<T>*>(ptr);
	CppObject<T>::Storage::destroy(src_obj->value);
}

/*     */

enum TypeConstantFlags {
	TYPE_METHOD_INIT = 1,
	TYPE_METHOD_COPY = 2,
	TYPE_METHOD_COMPARE = 4,
	TYPE_METHOD_TO_STRING = 8,
	TYPE_METHOD_TO_XML = 16,
	TYPE_METHOD_CHECK_INVARIANT = 32,
	TYPE_METHOD_ENUMERATE = 64,
	TYPE_METHOD_DESTROY = 128,
};

template <typename T>
struct TypeConstant {
	static Type make(const char * name,
		const Type* parent = &type_Object,
		int flags = TYPE_METHOD_INIT | TYPE_METHOD_COPY | TYPE_METHOD_DESTROY | TYPE_METHOD_TO_STRING | TYPE_METHOD_TO_XML)
	{
		//std::ios_base::fmtflags sflags = std::cout.flags();
		//std::cout << "Initialize type '" << name <<"', flags 0x";
		//std::cout.flags(sflags | std::ios_base::hex);
		//std::cout << flags << std::endl;
		//std::cout << "Initialize type '" << name <<"', use custom toXML:" << (flags&TYPE_METHOD_TO_XML) << std::endl;
		//std::cout.flags(sflags);
		::Type t = {
			name,
			sizeof(cpptesk::atl::CppObject<T>),
			parent,
			(flags & TYPE_METHOD_INIT)? init_CppObject<T> : &::init_Default,
			(flags & TYPE_METHOD_COPY)? copy_CppObject<T> : &::copy_Default,
			(flags & TYPE_METHOD_COMPARE)? compare_CppObject<T> : &::compare_Default,
			(flags & TYPE_METHOD_TO_STRING)? toString_CppObject<T> : &::to_string_Default,
			(flags & TYPE_METHOD_TO_XML)? toXML_CppObject<T> : &::to_XML_Default,
			(flags & TYPE_METHOD_CHECK_INVARIANT)? checkInvariant_CppObject<T> : &::check_invariant_Default,
			(flags & TYPE_METHOD_ENUMERATE)? enumerate_CppObject<T> : &::enumerate_Default,
			(flags & TYPE_METHOD_DESTROY)? destroy_CppObject<T> : &::destroy_Default,
		};
		return t;
	}
};

#define CPPTESK_TYPE_CONSTANT_INIT(type,name,parent) \
{										 \
	name,								 \
	sizeof(cpptesk::atl::CppObject<type>), \
	parent,								 \
	&cpptesk::atl::init_CppObject<type>,	 \
	&cpptesk::atl::copy_CppObject<type>,	 \
	&cpptesk::atl::compare_CppObject<type>,\
	&cpptesk::atl::toString_CppObject<type>,	\
	&cpptesk::atl::toXML_CppObject<type>,	 \
	&::check_invariant_Default,			 \
	&::enumerate_Default,				 \
	&cpptesk::atl::destroy_CppObject<type>			 \
}\

#ifdef _MSC_VER
#define CPPTESK_STRDUP ::_strdup
#else
#define CPPTESK_STRDUP ::strdup
#endif
//     ,  0 
template<>
struct _Storage<char *> {
	typedef char* storage_t;

	static char*& ref(storage_t& r) { return r; }
	static char* const& ref(const storage_t& r) { return *(char * const*)(&r); }
	static char** ptr(storage_t& r) { return &r; }
	static const char** ptr(const storage_t& r) { return (const char**)&r; }

	static storage_t empty() { return 0; }
	static void makeEmpty(storage_t * place) { *place = 0; }
	static bool isEmpty(const storage_t& val) { return val == 0; }
	static storage_t default_ref() { return new char(0); }
	static void destroy(storage_t& val) { delete val; val = 0; }
	static storage_t clone(const storage_t& src) { return CPPTESK_STRDUP(src); }

	//   
	static storage_t wrap_const(const char* src) { return (src==0)? 0 : CPPTESK_STRDUP(src); }
	//    .   
	//       destroy
	static storage_t wrap(char* src) { return src; }
	//   
	static storage_t fromNothing() { return new char(0); }
	//  .
	static void assign(storage_t& dst, const storage_t& src) {
		if (!isEmpty(dst) && !isEmpty(src)) dst = CPPTESK_STRDUP(src);
		else if (isEmpty(src)) destroy(dst);
		else if (isEmpty(dst)) dst = clone(src);
	}
};


	}
}

#endif // __cplusplus
#endif // CPPOBJECT_H
