/*
 * Copyright 2012-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK (cpptesk::examples)
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Family of scenarios for the MOESI cache coherence protocols:
//           (1) MSI   (Modified,                   Shared, Invalid)
//           (2) MOSI  (Modified, Owned,            Shared, Invalid)
//           (3) MESI  (Modified,        Exclusive, Shared, Invalid)
//           (4) MOESI (Modified, Owned, Exclusive, Shared, Invalid)
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <moesi_config.hpp>
#include <moesi_model.hpp>

#include <hw/state/state.hpp>
#include <ts/scenario.hpp>

using namespace cpptesk::ts;
using namespace cpptesk::hw;
using namespace cpptesk::hw::state;

namespace cpptesk {
namespace examples {
namespace moesi {

/**
 * Base scenario for verifying MOESI cache coherence protocols.
 */
SCENARIO(MoesiScenario)
{
public:
    ////////////////////////////////////////////////////////////////////////////
    // Abstraction techniques
    ////////////////////////////////////////////////////////////////////////////

    /// State abstraction techniques.
    enum StateAbstraction
    {
        /// An abstract state is a vector of cache line states (in all cores).
        /// Abstraction: none.
        /// Example: < M, S, S, I > => < M, S, S, I >.
        VECTOR,
        /// An abstract state is a multiset of cache line states (in all cores).
        /// Abstraction: state order abstraction (form of the symmetry abstraction).
        /// Example: < M, S, S, I > => { M, S, S, I }.
        MULTISET,
        /// An abstract state is a set of cache line states (in all cores).
        /// Abstraction: state order abstraction + state number abstraction.
        /// Example: < M, S, S, I > => { M, S, I }.
        SET
    };

    /// Stimulus abstraction techniques.
    enum StimulusAbstraction
    {
        /// An abstract stimulus is a pair < operation, core index >.
        /// Abstraction: none.
        /// Example: < M, S, S, I > -< write, 1 >-> < I, M, I, I >.
        INDEX,
        /// An abstract stimulus is a pair < operation, core state >.
        /// Abstraction: non-deterministic choice of a core with a given state.
        /// Example: { M, S, I } -< write, S >-> { M, I }.
        STATE
    };

    ///////////////////////////////////////////////////////////////////////////
    // Constructor / destructor
    ///////////////////////////////////////////////////////////////////////////

    /// Scenario constructor.
    MoesiScenario(): _model(0), _states_number(0)
    {
        setup("MOESI Scenario", &MoesiScenario::init, &MoesiScenario::finish, &MoesiScenario::get_state);

        // Read the test configuration.
        MoesiConfig::get().parse("test.cfg");

        const CacheProtocol protocol = CacheProtocol(MoesiConfig::get().cache_protocol);
        const size_t number_of_cores = size_t(MoesiConfig::get().cores_number);

        // Construct the model under test according to the parameters.
        _model = new MoesiModel(protocol, number_of_cores);

        // Register scenario methods.
        if(MoesiConfig::get().read_enable)  { SCENARIO_METHOD(MoesiScenario::read);  }
        if(MoesiConfig::get().write_enable) { SCENARIO_METHOD(MoesiScenario::write); }
        if(MoesiConfig::get().evict_enable) { SCENARIO_METHOD(MoesiScenario::evict); }
        if(MoesiConfig::get().reset_enable) { SCENARIO_METHOD(MoesiScenario::reset); }

        // Fill the array of protocol states.
        if(has_modified(protocol))  { _states[_states_number++] = MODIFIED;  }
        if(has_owned(protocol))     { _states[_states_number++] = OWNED;     }
        if(has_exclusive(protocol)) { _states[_states_number++] = EXCLUSIVE; }
        if(has_shared(protocol))    { _states[_states_number++] = SHARED;    }
        if(has_invalid(protocol))   { _states[_states_number++] = INVALID;   }
    }

    /// Scenario destructor.
    virtual ~MoesiScenario()
    {
        if(_model)
            { delete _model; }
    }

    virtual bool init(int argc, char **argv);
    virtual void finish();

    ///////////////////////////////////////////////////////////////////////////
    // Scenario methods
    ///////////////////////////////////////////////////////////////////////////

    typedef IntAbcCtx MoesiContext;

    bool read  (MoesiContext &ctx);
    bool write (MoesiContext &ctx);
    bool evict (MoesiContext &ctx);
    bool reset (MoesiContext &ctx);

    ////////////////////////////////////////////////////////////////////////////
    // State abstraction
    ////////////////////////////////////////////////////////////////////////////

    /// Returns the state depending on the abstraction type.
    virtual std::string get_state();

    ////////////////////////////////////////////////////////////////////////////
    // Stimulus abstraction
    ////////////////////////////////////////////////////////////////////////////

    /// Returns the number of stimuli. 
    size_t get_number() const;
    /// Returns the core identifier or -1.
    size_t get_core(int stimulus) const;

private:
    /// Returns the number of cores.
    size_t get_indices_number() const;
    /// Returns the core identifier.
    size_t get_core_by_index(int index) const;

    /// Returns the number of different states.
    size_t get_states_number() const;
    /// Returns the core identifier.
    size_t get_core_by_state(int state) const;

private:
    /// Model under test.
    MoesiModel *_model;
    /// Cache states.
    CacheState _states[5];
    /// Number of states.
    size_t _states_number;
};

}}} // cpptesk::examples::moesi

