/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <iostream>

#include <hw/model.hpp>
#include <hw/message.hpp>

using namespace cpptesk::hw;

////////////////////////////////////////////////////////////////////////////////
// Message Model
////////////////////////////////////////////////////////////////////////////////
enum COP {
    READ,
    WRITE
};

std::ostream& operator << (std::ostream &out, COP cop)
{
    switch(cop)
    {
    case READ:
        return out << "READ";
    default:
        return out << "WRITE";
    }
}

MESSAGE(MSG)
{
public:
    MSG()
    {
        ADD_FIELD(MSG::cop);
        ADD_FIELD_ARRAY(MSG::data);
    }
    
    virtual ~MSG() {}

    SUPPORT_CLONE(MSG)

    DECLARE_TYPED_FIELD(COP, cop, 10);
    DECLARE_FIELD_ARRAY(data, 10, 16);
};

////////////////////////////////////////////////////////////////////////////////
// Design Model
////////////////////////////////////////////////////////////////////////////////
MODEL(DUT)
{
public:
////////////////////////////////////////////////////////////////////////////////
// Input interface
////////////////////////////////////////////////////////////////////////////////
    DECLARE_IFACE(iface);

    DUT()
    {
        ADD_IFACE(iface);
	CALL_LOOP(process2);
    }

////////////////////////////////////////////////////////////////////////////////
// Design Stimuli
////////////////////////////////////////////////////////////////////////////////

    DECLARE_PROCESS(process2)
    {
        START_PROCESS();

        DEBUG_PRINT(DEBUG_USER, "1: time=" << std::dec << time());
        CYCLE();
	DEBUG_PRINT(DEBUG_USER, "2: time=" << std::dec << time());

	DEBUG_PRINT(DEBUG_ERROR, "Bug has been found in process2");

        STOP_PROCESS();
    }

    DECLARE_PROCESS(process1)
    {
        START_PROCESS();

	SET_PRIORITY(PRIORITY_99);

        CALL_PROCESS(PARALLEL, process2, GET_IFACE(), GET_MESSAGE());

        DEBUG_PRINTF(DEBUG_USER, "[%03d]: thread: stage 0", process.get_pid());
        DELAY(1);

        RAISE_EVENT("process1");
        DELAY(1);

        DEBUG_PRINTF(DEBUG_USER, "[%03d]: thread: stage 1", process.get_pid());
        WAIT(true || false);

        STOP_PROCESS();
    }

    DECLARE_STIMULUS(stimulus)
    {
        int i;
        START_STIMULUS(SEQUENTIAL);

        CAPTURE_IFACE();
	{
	    // No delay
	}
	RELEASE_IFACE();

        // Create a new process
        CALL_PROCESS(SEQUENTIAL, process1, GET_IFACE(), GET_MESSAGE());

        // Execute the code sequentially within the given process
        CALL_PROCESS_EX(PARALLEL, DUT::process1, GET_IFACE(), GET_MESSAGE(), START_THIS_CYCLE | HIGHEST_PRIORITY);

        WAIT_EVENT("process1");
        DEBUG_PRINT(DEBUG_USER, "An event has been received");

	// Just call it as a function
	process1(GET_PARAMETERS());

        for(i = 0; i < 9; i++)
        {
            DEBUG_PRINTF(DEBUG_USER, "[%03d]: stage=%d", process.get_pid(), i);
            CYCLE();
        }

        DEBUG_PRINTF(DEBUG_USER, "[%03d]: stage=%d", process.get_pid(), i);

	SEND_REACTION(PARALLEL, GET_IFACE(), GET_MESSAGE());
        CYCLE();

        STOP_STIMULUS();
    }

////////////////////////////////////////////////////////////////////////////////
// Design Mediators
////////////////////////////////////////////////////////////////////////////////
    virtual void onEveryCycle()
    {
        DEBUG_PRINT(DEBUG_USER, "onEveryCycle: time=" << std::dec << time());
    }

    virtual bool isExitCondition() { return time() == 100; }
};

////////////////////////////////////////////////////////////////////////////////
// Testbench
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    DUT dut;
    MSG msg;

    SET_DEBUG_LEVEL(DEBUG_INFO, true);
    SET_DEBUG_LOGFILE("cpptesk.log");

    SET_DEBUG_STYLE(DEBUG_USER, BLACK, CYAN, CYAN, NORMAL);

    dut.initialize(argc, argv);

    for(int i = 0; i < 100; i++)
    {
        msg.randomize();
	DEBUG_PRINT(DEBUG_USER, "Start stimulus");
        CALL_STIMULUS_OF(dut, DUT::stimulus, dut.iface, msg);
        dut.cycle();
    }

    for(int i = 0; i < 2000; i++)
    {
        dut.quantum();
    }

    DEBUG_PRINT(DEBUG_USER,  "CPPTESKHW_VERSION=" <<  CPPTESKHW_VERSION);
    DEBUG_PRINT(DEBUG_ERROR, "This is the end");

    dut.finalize();

    return 0;
}

