/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Mikhail Chupilko (chupilko@ispras.ru)
// SYNOPSIS: Reaction arbiters
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/arbiter/base.hpp>

namespace cpptesk {
namespace hw {

////////////////////////////////////////////////////////////////////////////////
// Adaptive Reaction Arbiter
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_COMPARE_ARBITER ::cpptesk::hw::AdaptiveArbiter

#ifdef HW_SHORT_NAMES
    #define COMPARE_ARBITER CPPTESK_COMPARE_ARBITER
#endif // HW_SHORT_NAMES

class AdaptiveArbiter : public Arbiter
{
public:
    AdaptiveArbiter() {}
    virtual ~AdaptiveArbiter() {}

    virtual void appendRequest(const Process &process, const Message &message)
    {
        _requests.push_back(ArbiterRequest(process, message));
    }

    virtual bool chooseRequest(const Process &process, const Message &message)
    {
        CPPTESK_ASSERTION(!_requests.empty(), "There are no requests to choose from");
        SharedPtr<Message> message_copy(message.clone());
        message_copy->validateFields();

        for(requests_t::iterator i = _requests.begin(); i != _requests.end(); i++)
        {
            SharedPtr<Message> kept_message_copy(i->message->clone());
            kept_message_copy->validateFields();

            if(i->process == &process && kept_message_copy->compare(*message_copy) == Message::COMPARE_OK)
            {
                return true;
            }
        }

        return false;
    }

    virtual void removeRequest(const Process &process, const Message &message)
    {
        CPPTESK_ASSERTION(!_requests.empty(), "There are no requests to remove");
        SharedPtr<Message> message_copy(message.clone());
        message_copy->validateFields();

        int first_non_optional = -1;
        int first_optional = -1;
        requests_t::iterator i;
        int number;

        for(i = _requests.begin(), number = 0; i != _requests.end(); i++, number++)
        {
            SharedPtr<Message> kept_message_copy(i->message->clone());
            kept_message_copy->validateFields();

            if(i->process == &process && kept_message_copy->compare(*message_copy) == Message::COMPARE_OK)
            {
                if(kept_message_copy->isOptional())
                {
                    if(first_optional == -1)
                        first_optional = number;
                }
                else
                {
                    if(first_non_optional == -1)
                        first_non_optional = number;
                }
            }
        }

        CPPTESK_ASSERTION(first_optional != -1 || first_non_optional != -1,
                          "There should be either optional or non-optional message");

        if(first_non_optional == -1)
            first_non_optional = first_optional;

        for(i = _requests.begin(), number = 0; number < first_non_optional; i++, number++);
        _requests.erase(i);
    }

    virtual bool isModelBased() const { return false; }
};

}} // namespace cpptesk::hw

