/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Mikhail Chupilko (chupilko@ispras.ru)
// SYNOPSIS: Reaction arbiters
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/arbiter/base.hpp>

namespace cpptesk {
namespace hw {

////////////////////////////////////////////////////////////////////////////////
// Two-Level Reaction Arbiter
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_TWO_LEVEL_ARBITER ::cpptesk::hw::TwoLevelArbiter

#ifdef HW_SHORT_NAMES
    #define TWO_LEVEL_ARBITER CPPTESK_TWO_LEVEL_ARBITER
#endif // HW_SHORT_NAMES

template <class PrimaryArbiter, class SecondaryArbiter>
  class TwoLevelArbiter : public Arbiter
{
public:
    TwoLevelArbiter()
    {
        CPPTESK_ASSERTION(_arb1.isModelBased() && !_arb2.isModelBased(), "[SYSTEM] TwoLevelArbiter: A1 should be model based and A2 should be adaptive!");
    }
    virtual ~TwoLevelArbiter() {}

    virtual void appendRequest(const Process &process, const Message &message)
    {
        _arb1.appendRequest(process, message);
    }

    /// Add request to the list of requests passed first stage
    virtual bool preChooseRequest(const Process &process, const Message &message)
    {
        if (_arb1.chooseRequest(process, message))
        {
            _requests.push_back(ArbiterRequest(process, message));
        }
        else
            return false;

        return true;
    }

    virtual bool chooseRequest(const Process &process, const Message &message)
    {
        CPPTESK_ASSERTION(!_requests.empty(), "No request has been chosen by the first arbiter!");

        for(requests_t::iterator i = _requests.begin(); i != _requests.end(); i++)
        {
            _arb2.appendRequest(*(i->process), *(i->message));
        }

        /// Ask arb2 to find our request in the list of pre-selected requests
        bool result = _arb2.chooseRequest(process, message);

        /// Remove all requests from arb2
        for(requests_t::iterator i = _requests.begin(); i != _requests.end(); i++)
        {
            _arb2.removeRequest(*(i->process), *(i->message));
        }

        /// Remove our request from the list of pre-selected requests
        for(requests_t::iterator i = _requests.begin(); i != _requests.end(); i++)
        {
            if(i->process == &process)
            {
                _requests.erase(i);
                break;
            }
        }

        return result;
    }

    virtual void removeRequest(const Process &process, const Message &message)
    {
        _arb1.removeRequest(process, message);
    }

    virtual bool isModelBased() const { return _arb1.isModelBased() && _arb2.isModelBased(); }

    virtual requests_t getRequests() const { return _arb1.getRequests(); }
    virtual requests_t::iterator begin() { return _arb1.begin(); }
    virtual requests_t::iterator end() { return _arb1.end(); }

protected:
    PrimaryArbiter _arb1;
    SecondaryArbiter _arb2;
};

}} // namespace cpptesk::hw

