/** @file autofsm.hpp
 * This file contains automatical FSM generation class description and implementation.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#include <algorithm>
#include <vector>

#include "ts/scenario.hpp"
#include "hw/model.hpp"

namespace cpptesk {
namespace hw {

enum ShowMode { SHOW_NUMBER, SHOW_LINE, SHOW_ALL, SHOW_NONE };

/**
 * This class provides functionality of automatical generation of FSM from models.
 */
template<class UserModel>
class AutoFSM
{
public:
  /**
   * Constructor with reference to DUT model */
  AutoFSM(UserModel &dut) : dut(dut) { _show_start = SHOW_ALL; _show_cycle = SHOW_ALL; _show_guard = SHOW_ALL; };
  /**
   * Default virtual destructor */
  virtual ~AutoFSM() {};
  /**
   * Set parameter show_start */
  void set_show_start(uint show_start) { _show_start = show_start; };
  /**
   * Set parameter show_cycle */
  void set_show_cycle(uint show_cycle) { _show_cycle = show_cycle; };
  /**
   * Set parameter show_guard */
  void set_show_guard(uint show_guard) { _show_guard = show_guard; };
  
  /**
   * Initialize AutoFSM */
  void initialize() { _initial_state = get_state(); };
  
  /**
   * Return current state */
  std::string get_state()
  {
    const std::vector<Process::PointDescriptor> points = dut.getPipelineSlice();
    std::vector<int> start_lines, cycle_lines, guard_lines;
    
    std::stringstream out;
    
    int start_counter = 0, cycle_counter = 0, guard_counter = 0;
    
    for (unsigned int i = 0; i < points.size(); i++)
    {
        Process::PointDescriptor point = points[i];
    
        switch (point.type)
        {
            case START: if(_show_start == SHOW_ALL) { out << "{" << point.name << "}"; } else if(_show_start == SHOW_LINE) { start_lines.push_back(point.line); } else { start_counter++; } break;
            case CYCLE: if(_show_cycle == SHOW_ALL) { out << "{" << point.name << "}"; } else if(_show_cycle == SHOW_LINE) { cycle_lines.push_back(point.line); } else { cycle_counter++; } break;
            case GUARD: if(_show_guard == SHOW_ALL) { out << "{" << point.name << "}"; } else if(_show_guard == SHOW_LINE) { guard_lines.push_back(point.line); } else { guard_counter++; } break;
            default: CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_INFO, "Unknown type of point " <<  point.type << " in " << point.file << ":" << point.line);
        }
    }

    if(_show_start == SHOW_LINE)
    {
        sort(start_lines.begin(), start_lines.end());
	for(std::vector<int>::iterator i = start_lines.begin(); i != start_lines.end(); i++)
            if(i == start_lines.begin() || *i != *(i - 1))
                out << "{" << *i << "}";
            else
                out << "x";
    }
    if(_show_cycle == SHOW_LINE)
    {
        sort(cycle_lines.begin(), cycle_lines.end());
        for(std::vector<int>::iterator i = cycle_lines.begin(); i != cycle_lines.end(); i++)
            if(i == cycle_lines.begin() || *i != *(i - 1))
                out << "{" << *i << "}";
            else
                out << "x";
    }
    if(_show_guard == SHOW_LINE)
    {
        sort(guard_lines.begin(), guard_lines.end());
        for(std::vector<int>::iterator i = guard_lines.begin(); i != guard_lines.end(); i++)
            if(i == guard_lines.begin() || *i != *(i - 1))
                out << "{" << *i << "}";
            else
                out << "x";
    }

    if(_show_start == SHOW_NUMBER) out << "{" << start_counter << "}";
    if(_show_cycle == SHOW_NUMBER) out << "{" << cycle_counter << "}";
    if(_show_guard == SHOW_NUMBER) out << "{" << guard_counter << "}";

    CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_MORE, "get_state(): " << out.str());
    return out.str();
  };

  void scen_nop()
  {
      int i = 0;
      std::string state = get_state();
      dut.cycle();
      if(state != _initial_state) while(state == get_state() && (i++ < 100)) dut.cycle(); // TODO: how is it possible to differentiate stable and unstable states?
  };

protected:
  /**
   * Show labels START or only their number */
  uint _show_start;
  /**
   * Show labels CYCLE or only their number */
  uint _show_cycle;
  /**
   * Show labels WAIT or only their number */
  uint _show_guard;
  /**
   * Reference to the DUT model */
  UserModel &dut;
  /**
   * State which is right after DUT reboot */
  std::string _initial_state;
};

}}
