/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Pair of Reactions
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/reaction_pair.hpp>
#include <hw/diagnostics/rules/rule.hpp>
#include <hw/interface.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

ReactionPair::ReactionPair()
{
    _group = -1;
    _collapsed = false;
}

ReactionPair::ReactionPair(Reaction &impl_reaction,
                           Reaction &spec_reaction,
                           Reaction &collapsed_reaction):
        _impl_reaction(impl_reaction),
        _spec_reaction(spec_reaction),
        _collapsed_reaction(collapsed_reaction)
{
    _group = -1;
    _collapsed = false;
}

void ReactionPair::set_collapsed(bool collapsed)
{
    _collapsed = collapsed;
}

bool ReactionPair::is_collapsed()
{
    return _collapsed;
}

void ReactionPair::set_group(int group)
{
    _group = group;
}

int  ReactionPair::get_group()
{
    return _group;
}

void ReactionPair::add_impl_reaction(Reaction &impl_reaction)
{
    _impl_reaction = impl_reaction;
}

void ReactionPair::add_spec_reaction(Reaction &spec_reaction)
{
    _spec_reaction = spec_reaction;
}

void ReactionPair::add_collapsed_reaction(Reaction &collapsed_reaction)
{
    _collapsed_reaction = collapsed_reaction;
}

Reaction& ReactionPair::get_impl_reaction()
{
    return _impl_reaction;
}

Reaction& ReactionPair::get_spec_reaction()
{
    return _spec_reaction;
}

Reaction& ReactionPair::get_collapsed_reaction()
{
    return _collapsed_reaction;
}

void ReactionPair::process(cycle_t reaction_timeout, cycle_t unexpected_reaction_timeout)
{
    _error = 0;

    if(Rule::compare(_collapsed_reaction.getSpecMessage(),
                     _collapsed_reaction.getImplMessage()) != EXACT_MATCH
                     && _collapsed_reaction.getError() != MISSING_REACTION 
                     && _collapsed_reaction.getError() != UNEXPECTED_REACTION)
        _error |= DATA_ERROR;
    
    if(_collapsed && !(_collapsed_reaction.getError() == MISSING_REACTION || 
       _collapsed_reaction.getError() == UNEXPECTED_REACTION))
    {
        // Compare the iface
        if(_spec_reaction.where() != _impl_reaction.where())
            _error |= IFACE_ERROR;

        // Compare the time
        const ::cpptesk::hw::Interface *iface_impl = _impl_reaction.where();
        cycle_t impl_time_reaction;
        if(iface_impl->isUnexpectedReactionTimeoutValid())
        {
            impl_time_reaction = iface_impl->getUnexpectedReactionTimeout();    
        }
        else
            impl_time_reaction = unexpected_reaction_timeout;

        const ::cpptesk::hw::Interface *iface_spec = _spec_reaction.where();
        cycle_t spec_time_reaction;
        if(iface_spec->isReactionTimeoutValid())
        {
            spec_time_reaction = iface_spec->getReactionTimeout();
        }
        else
            spec_time_reaction = reaction_timeout;
        
        int temp_time = _spec_reaction.whenStarted() - impl_time_reaction;
        cycle_t min_time_reaction = temp_time < 0 ? 0 : temp_time;
        cycle_t max_time_reaction = _spec_reaction.whenStarted() + spec_time_reaction;
        
        if(min_time_reaction > _impl_reaction.whenStarted() || 
        max_time_reaction < _impl_reaction.whenStarted())
            _error |= TIME_ERROR;
    }

    _group = _error;
}

}}}
