/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Recombination of INCORRECT reaction pairs
//           Rule #2: {[y1:x2]+[x1:y2] => [y1:y2]+[x1:x2]}
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/rules/rule2.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

int Rule2::check(cpptesk::hw::Reaction &r1, cpptesk::hw::Reaction &r2)
{
    if(r1.getError() == INCORRECT_REACTION && r2.getError() == INCORRECT_REACTION &&
       r1.getSpecMessage() != NULL && r2.getSpecMessage() != NULL &&
       r1.getSpecMessage()->getName() == r2.getSpecMessage()->getName())
    {
        int metric_aa = compare(r1.getSpecMessage(), r1.getImplMessage());
        int metric_ab = compare(r1.getSpecMessage(), r2.getImplMessage());
        int metric_ba = compare(r2.getSpecMessage(), r1.getImplMessage());
        int metric_bb = compare(r2.getSpecMessage(), r2.getImplMessage());

        if((metric_ab < metric_aa && metric_ab < metric_bb) ||
           (metric_ba < metric_aa && metric_ba < metric_bb))
            return (metric_ab < metric_ba ? metric_ab : metric_ba);
    }

    return FALSE_MATCH;
}

int Rule2::check_completely(cpptesk::hw::Reaction &r1, cpptesk::hw::Reaction &r2)
{
    if(r1.getError() == INCORRECT_REACTION && r2.getError() == INCORRECT_REACTION &&
           r1.getSpecMessage()->getName() == r2.getSpecMessage()->getName())
    {
        int metric_aa = compare(r1.getSpecMessage(), r1.getImplMessage());
        int metric_ab = compare(r1.getSpecMessage(), r2.getImplMessage());
        int metric_ba = compare(r2.getSpecMessage(), r1.getImplMessage());
        int metric_bb = compare(r2.getSpecMessage(), r2.getImplMessage());

        if((metric_ab < metric_aa && metric_ab < metric_bb) ||
           (metric_ba < metric_aa && metric_ba < metric_bb))
            return metric_ab + metric_ba;
    }

    return FALSE_MATCH;
}

ReactionUnitVector Rule2::apply(ReactionUnit &ru1, ReactionUnit &ru2)
{
    cpptesk::hw::Reaction &r1 = ru1.getReaction();
    cpptesk::hw::Reaction &r2 = ru2.getReaction();

    // Take min "start" and max "end" time of reactions
    cycle_t start_time = r1.whenStarted() < r2.whenStarted() ? r1.whenStarted() : r2.whenStarted();
    cycle_t end_time   = r1.whenStopped() > r2.whenStopped() ? r1.whenStopped() : r2.whenStopped();

    bool normal1 = (compare(r1.getSpecMessage(), r2.getImplMessage()) == EXACT_MATCH);
    bool normal2 = (compare(r2.getSpecMessage(), r1.getImplMessage()) == EXACT_MATCH);

    cpptesk::hw::Reaction new_r1 = Reaction(normal1 ? NORMAL_REACTION : INCORRECT_REACTION,
                                            *(r1.where()), *(r1.getSpecMessage()), *(r2.getImplMessage()), start_time, end_time,
                                            normal1 ? "normal reaction after rule#2" : "incorrect reaction after rule#2");
    cpptesk::hw::Reaction new_r2 = Reaction(normal2 ? NORMAL_REACTION : INCORRECT_REACTION,
                                            *(r2.where()), *(r2.getSpecMessage()), *(r1.getImplMessage()), start_time, end_time,
                                            normal2 ? "normal reaction after rule#2" : "incorrect reaction after rule#2");

    ReactionUnitVector result = ReactionUnitVector(new_r1, new_r2, ru1.get_id(), ru2.get_id(), ru2.get_id(), ru1.get_id());

    ru1.set_collapsed();
    ru2.set_collapsed();

    return result;
}

std::string Rule2::toString()
{
    return "for INCORRECT and INCORRECT reactions: {[a:b]+[b:a] => [a:a]+[b:b]}";
}

}}}
