/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Recombination of INCORRECT and MISSING reaction pairs
//           Rule #4: {[a:b]+[b:x] => [b:b]+[a:x]}
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/rules/rule4.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

int Rule4::check(cpptesk::hw::Reaction &r1, cpptesk::hw::Reaction &r2)
{
    if(r1.getSpecMessage() != NULL && r2.getSpecMessage() != NULL &&
       r1.getSpecMessage()->getName() == r2.getSpecMessage()->getName())
    {
        if((r1.getError() == INCORRECT_REACTION && r2.getError() == MISSING_REACTION) ||
           (r2.getError() == INCORRECT_REACTION && r1.getError() == MISSING_REACTION))
        {
            cpptesk::hw::Reaction &in_r1 = (r1.getError() == INCORRECT_REACTION ? r1 : r2);
            cpptesk::hw::Reaction &in_r2 = (r1.getError() == INCORRECT_REACTION ? r2 : r1);

            int metric_aa = compare(in_r1.getSpecMessage(), in_r1.getImplMessage());
            int metric_ba = compare(in_r2.getSpecMessage(), in_r1.getImplMessage());

            return (metric_ba < metric_aa ? metric_ba : WORST_MATCH);
        }
    }

    return FALSE_MATCH;
}

ReactionUnitVector Rule4::apply(ReactionUnit &ru1, ReactionUnit &ru2)
{
    cpptesk::hw::Reaction &r1 = ru1.getReaction();
    cpptesk::hw::Reaction &r2 = ru2.getReaction();

    bool is_r1_incorrect = r1.getError() == INCORRECT_REACTION;

    const cpptesk::hw::Message &m1 = *(is_r1_incorrect ? r2.getSpecMessage() : r1.getSpecMessage()); // for new incorrect
    const cpptesk::hw::Message &m2 = *(is_r1_incorrect ? r1.getImplMessage() : r2.getImplMessage()); // for new incorrect
    const cpptesk::hw::Message &m3 = *(is_r1_incorrect ? r1.getSpecMessage() : r2.getSpecMessage()); // for new missing

    int id1 = (is_r1_incorrect ? ru2.get_id() : ru1.get_id());
    int id2 = (is_r1_incorrect ? ru1.get_id() : ru2.get_id());

    // Take min "start" and max "end" time of reactions
    cycle_t start_time = r1.whenStarted() < r2.whenStarted() ? r1.whenStarted() : r2.whenStarted();
    cycle_t end_time   = r1.whenStopped() > r2.whenStopped() ? r1.whenStopped() : r2.whenStopped();

    bool normal = (compare(m1, m2) == EXACT_MATCH);
    cpptesk::hw::Reaction new_r1 = Reaction(normal ? NORMAL_REACTION : INCORRECT_REACTION, *(r1.where()), m1, m2,
                                            start_time, end_time,
                                            normal ? "normal reaction after rule#4" : "incorrect reaction after rule#4");
    cpptesk::hw::Reaction new_r2 = Reaction(MISSING_REACTION, *(r2.where()), m3,
                                            start_time, end_time, "missing reaction after rule#4");

    ReactionUnitVector result = ReactionUnitVector(new_r1, new_r2, id1, id2, id2, -1);

    ru1.set_collapsed();
    ru2.set_collapsed();

    return result;
}

std::string Rule4::toString()
{
    return "for INCORRECT and MISSING reactions: {[a:b]+[b:x] => [b:b]+[a:x]}";
}

}}}
