/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Finite state machines
///////////////////////////////////////////////////////////////////////////////

#include <hw/fsm/fsm.hpp>

namespace cpptesk {
namespace hw {
namespace fsm {

void BaseFsm::add_transition(const _State &source, const _Stimulus &stimulus, action_t action_method, const _State &destination, const std::string &text)
{
    CPPTESK_ASSERTION(!_compiled, "The FSM transitions have been already compiled");

    _Action action(*this, action_method);
    _Transition transition(source, stimulus, action, destination, text);

    _transitions.push_back(transition);
}

void BaseFsm::_compile()
{
    CPPTESK_ASSERTION(!_compiled, "The FSM transitions have been already compiled");

    // Construct a preliminary transition table and calculates the number of states.
    std::map<state_t, _stimulus_map_t> transition_table;
    _transitions_t generic_transitions;

    for(_transitions_t::iterator i = _transitions.begin(); i != _transitions.end(); i++)
    {
        const state_t    q = i->state().id();
        const stimulus_t x = i->stimulus().id();
        const state_t    n = i->next().id();

        if(q != ANY_STATE)
            { _states[q] = i->state(); }
        if(x != ANY_STIMULUS)
            { _stimuli[x] = i->stimulus(); }
        if(n != ANY_STATE)
            { _states[n] = i->next(); }

        if(q != ANY_STATE && x != ANY_STIMULUS)
        {
            CPPTESK_ASSERTION(transition_table[q].find(x) == transition_table[q].end(),
                "The FSM has nondeterministic behavior: " << transition_table[q][x]);

            transition_table[q][x] = *i;
        }
        else
        {
            generic_transitions.push_back(*i);
        }
    }

    for(_transitions_t::iterator i = generic_transitions.begin(); i != generic_transitions.end(); i++)
    {
        const state_t    q = i->state().id();
        const stimulus_t x = i->stimulus().id();
        const state_t    n = i->next().id();

        for(_states_t::iterator  qi = _states.begin();  qi != _states.end();  qi++)
        for(_stimuli_t::iterator xi = _stimuli.begin(); xi != _stimuli.end(); xi++)
        {
            const state_t    qq = qi->first;
            const stimulus_t xx = xi->first;
            const state_t    nn = (n == ANY_STATE ? qq : n);

            if((q == ANY_STATE || q == qq) && (x == ANY_STIMULUS || x == xx))
            {
                _State      state      (qq, i->state().text());
                _Stimulus   stimulus   (xx, i->stimulus().text());
                _State      next       (nn, i->next().text());

                _Transition transition (state, stimulus, i->action(), i->post_action(), next, i->text());

                if(transition_table[qq].find(xx) == transition_table[qq].end())
                    { transition_table[qq][xx] = transition; }
            }
        }
    }

    // Set the size of the transition table.
    _transition_table.resize(transition_table.size());

    // Fill the transition table.
    for(std::map<state_t, _stimulus_map_t>::iterator i = transition_table.begin(); i != transition_table.end(); i++)
    {
        _transition_table.put(i->first, i->second);
    }

    _compiled = true;
}

BaseFsm::_Transition& BaseFsm::_get_transition(state_t state_id, stimulus_t stimulus_id)
{
    _transition_table_t::iterator q = _transition_table.find(state_id);
    CPPTESK_ASSERTION(q != _transition_table.end(), "The state id " << state_id <<" is unknown");

    _stimulus_map_t::iterator x = q->find(stimulus_id);
    CPPTESK_ASSERTION(x != q->end(), "The stimulus id " << stimulus_id << " is unknown");

    return x->second;
}

}}} // namespace cpptesk::hw::fsm

