/*
 * Copyright 2012-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Finite state machines
// ///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/message.hpp>
#include <hw/utils/hash_table.hpp>
#include <hw/utils/shared_ptr.hpp>
#include <hw/unused.hpp>

#include <list>
#include <map>
#include <string>

#define CPPTESK_FSM(UserFsm) \
    class UserFsm: public ::cpptesk::hw::fsm::BaseFsm

#define CPPTESK_DECLARE_ACTION(action_method) \
    void action_method(::cpptesk::hw::fsm::Transition &transition)

#define CPPTESK_DEFINE_ACTION(action_method) \
    void action_method(::cpptesk::hw::fsm::Transition &transition)

#define CPPTESK_GET_TRANSITION() transition

#define CPPTESK_INITIAL_STATE(state_id) \
    set_initial_state(state_id)

#define CPPTESK_TRANSITION_WITH_COMMENT(state_id, stimulus_id, action_method, next_id, text) \
    do { \
        ::cpptesk::hw::fsm::BaseFsm::_State state(state_id, #state_id); \
        ::cpptesk::hw::fsm::BaseFsm::_Stimulus stimulus(stimulus_id, #stimulus_id); \
        ::cpptesk::hw::fsm::BaseFsm::_State next(next_id, #next_id); \
        add_transition(state, stimulus, (::cpptesk::hw::fsm::BaseFsm::action_t)&action_method, next, text); \
    } while(false)

#define CPPTESK_TRANSITION(state_id, stimulus_id, action_method, next_id) \
    CPPTESK_TRANSITION_WITH_COMMENT(state_id, stimulus_id, action_method, next_id, "")

#ifdef HW_SHORT_NAMES
    #define FSM                     CPPTESK_FSM
    #define DECLARE_ACTION          CPPTESK_DECLARE_ACTION
    #define DEFINE_ACTION           CPPTESK_DEFINE_ACTION
    #define GET_TRANSITION          CPPTESK_GET_TRANSITION
    #define INITIAL_STATE           CPPTESK_INITIAL_STATE
    #define TRANSITION_WITH_COMMENT CPPTESK_TRANSITION_WITH_COMMENT
    #define TRANSITION              CPPTESK_TRANSITION
#endif // HW_SHORT_NAMES

namespace cpptesk {
namespace hw {
namespace fsm {

/**
 * A basic class for specifying deterministic finite state machines (FSMs).
 */
class BaseFsm
{
protected:
    class _State;
    class _Stimulus;
    class _Transition;

public:
    /// User state identifiers should be non-negative.
    typedef int state_t;
    /// User stimulus identifiers should be non-negative.
    typedef int stimulus_t;

    static const int $ANY = -1;

    static const state_t    ANY_STATE    = $ANY;
    static const stimulus_t ANY_STIMULUS = $ANY;

    /// FSM step descriptor.
    CPPTESK_MESSAGE(Step)
    {
        friend class BaseFsm;
	friend std::ostream& operator << (std::ostream &out, const BaseFsm::Step &step);

    public:
        CPPTESK_SUPPORT_CLONE(Step);

        Step(const Message &message, const _Transition &transition):
            _message(message.clone()), _transition(transition), _executed(false) {}

        /// Return the start identifier of the transition.
        state_t state() const { return _transition.state().id(); }
        /// Returns the stimulus identifier of the transition.
        stimulus_t stimulus() const { return _transition.stimulus().id(); }
        /// Return the next state identifier of the transition.
        state_t next() const { return _transition.next().id(); }

        /// Returns the message associated with the step.
        Message& message() { return *_message; }
        const Message &message() const { return *_message; }

    private:
        void start_action()
        {
            _Transition &tr = const_cast<_Transition &>(_transition);
            tr.start_action(*this);
        }

        const _State& stop_action()
        {
            CPPTESK_ASSERTION(!_executed, "The FSM step has been already executed");

            _Transition &tr = const_cast<_Transition &>(_transition);
            tr.stop_action(*this);
            _executed = true;

            return tr.next();
        }

        const _Transition& transition() const { return _transition; }

	bool is_executed() const { return _executed; }

    private:
        SharedPtr<Message> _message;
        const _Transition &_transition;

        bool _executed;
    };

    typedef void (BaseFsm::*action_t)(Step &);

protected:
    ///////////////////////////////////////////////////////////////////////////
    // FSM state
    ///////////////////////////////////////////////////////////////////////////
    class _State
    {
    public:
        _State(state_t id, const std::string &text):
            _id(id), _text(text) {}
        _State():
	    _id(-1), _text("<invalid-state>") {}
	~_State() {}

        state_t id() const { return _id; }
        const std::string& text() const { return _text; }

    private:
        state_t _id;
        std::string _text;
    };

    ///////////////////////////////////////////////////////////////////////////
    // FSM stimulus
    ///////////////////////////////////////////////////////////////////////////
    class _Stimulus
    {
    public:
        _Stimulus(stimulus_t id, const std::string &text):
	    _id(id), _text(text) {}
        _Stimulus():
            _id(-1), _text("<invalid-stimulus>") {}
        ~_Stimulus() {}

        stimulus_t id() const { return _id; }
        const std::string& text() const { return _text; }

    private:
        stimulus_t _id;
        std::string _text;
    };

    ///////////////////////////////////////////////////////////////////////////
    // FSM action
    ///////////////////////////////////////////////////////////////////////////
    class _Action
    {
    public:
        _Action(BaseFsm &fsm, action_t action):
            _fsm(&fsm), _action(action) {}
        _Action():
	    _fsm(0), _action(&BaseFsm::silent) {}
        ~_Action() {}

	void run(Step &step)
	{
	    if(_fsm) { (_fsm->*_action)(step); }
        }

    private:
        BaseFsm* _fsm;
	action_t _action;
    };

    ///////////////////////////////////////////////////////////////////////////
    // FSM transition
    ///////////////////////////////////////////////////////////////////////////
    class _Transition
    {
    public:
        _Transition(const _State &state, const _Stimulus &stimulus, const _Action &action, const _Action &post_action, const _State &next, const std::string &text):
            _state(state), _stimulus(stimulus), _action(action), _post_action(post_action), _next(next), _text(text) {}
        _Transition(const _State &state, const _Stimulus &stimulus, const _Action &action, const _State &next, const std::string &text):
            _state(state), _stimulus(stimulus), _action(action), _next(next), _text(text) {}
        _Transition():
            _text("<invalid-transition>") {}

        const _State& state() const { return _state; }
        const _Stimulus& stimulus() const { return _stimulus; }
        const _Action& action() { return _action; }
        const _Action& post_action() { return _post_action; }
        const _State& next() const { return _next; }
        const std::string& text() const { return _text; }

        void start_action(Step &step) { _action.run(step); }
	void stop_action(Step &step)  { _post_action.run(step); }

    private:
        _State      _state;
        _Stimulus   _stimulus;
        _Action     _action;
        _Action     _post_action;
        _State      _next;
        std::string _text;
    };

    friend std::ostream& operator << (std::ostream &out, const _State &state);
    friend std::ostream& operator << (std::ostream &out, const _Stimulus &stimulus);
    friend std::ostream& operator << (std::ostream &out, const _Transition &transition);

    ///////////////////////////////////////////////////////////////////////////
    // FSM itself
    ///////////////////////////////////////////////////////////////////////////
    BaseFsm():
        _compiled(false), _started(false), _initiated(false), _transition_table(100) {}

    virtual ~BaseFsm() {}

public:
    /// Initializes the FSM execution (sets the current state).
    void start(state_t state_id)
    {
        CPPTESK_ASSERTION(!_started, "The FSM execution has been already started");

        if(!_compiled)
	    { _compile(); }

        _state_id = state_id;
        _started = true;
    }

    /// Initializes the FSM execution.
    void start()
    {
        CPPTESK_ASSERTION(_initiated, "The initial state of the FSM is unknown");

	start(_initial_state_id);
    }

    /// Sets the initial state of the FSM.
    void set_initial_state(state_t initial_state_id)
    {
        _initial_state_id = initial_state_id;
	_initiated = true;
    }

    /// Finilizes the FSM execution.
    void stop() { _started = false; }

    /// Returns the current state of the FSM execution.
    state_t state() const { return _state_id; }
    /// Sets the current state of the FSM execution.
    void state(state_t state_id) { _state_id = state_id; }

    /// Returns the string representation of the current state.
    const std::string& state_text() const
    {
        _states_t::const_iterator i = _states.find(_state_id);
        return i->second.text();
    }

    /// Starts the transition (executes the action, but does not change the state).
    Step start_step(stimulus_t stimulus_id, const Message &message = CPPTESK_DEFAULT_MESSAGE)
    {
        CPPTESK_ASSERTION(_started, "The FSM execution has not been initialized");

        Step step(message, _get_transition(_state_id, stimulus_id));
        step.start_action();

        return step;
    }

    /// Commits the transition (changes the state).
    void stop_step(Step &step)
    {
        const _State &state = step.stop_action();
        _state_id = state.id();
    }

    /// Performs one step of the FSM execution.
    void step(stimulus_t stimulus_id)
    {
        Step step = start_step(stimulus_id);
        stop_step(step);
    }

    bool precondition(stimulus_t stimulus_id)
    {
        _transition_table_t::iterator q = _transition_table.find(_state_id);

        if(q == _transition_table.end())
            { return false; }

        _stimulus_map_t::iterator x = q->find(stimulus_id);

        return x != q->end();
    }

protected:
    /// Silent action, which is used as a default action.
    void silent(Step &) {}

    /// Adds the transition into the FSM.
    void add_transition(const _State &source, const _Stimulus &stimulus, action_t action_method, const _State &destination, const std::string &text);

private:
    typedef std::list<_Transition>              _transitions_t;

    typedef std::map<state_t, _State>           _states_t;
    typedef std::map<stimulus_t, _Stimulus>     _stimuli_t;
    typedef std::map<stimulus_t, _Transition>   _stimulus_map_t;
    
    typedef HashTable<state_t, _stimulus_map_t> _transition_table_t;

    /// Compiles the FSM transitions into the optimized internal format.
    void _compile();
    /// Returns the transition for the given state and stimulus.
    _Transition& _get_transition(state_t state_id, stimulus_t stimulus_id);

private:
    /// Transition table has been compiled.
    bool                _compiled;
    /// Exection has been started.
    bool                _started;
    /// Initial state has been set.
    bool                _initiated;

    state_t             _state_id;
    state_t             _initial_state_id;

    _transitions_t      _transitions;

    _states_t           _states;
    _stimuli_t          _stimuli;

    _transition_table_t _transition_table;
    
};

typedef BaseFsm::Step Transition;

inline std::ostream& operator << (std::ostream &out, const BaseFsm::_State &state)
{
    return out << state.text();
}

inline std::ostream& operator << (std::ostream &out, const BaseFsm::_Stimulus &stimulus)
{
    return out << stimulus.text();
}

inline std::ostream& operator << (std::ostream &out, const BaseFsm::_Transition &transition)
{
    out << "(";
    out << transition.state() << " -[";
    out << transition.stimulus() << "]-> ";
    out << transition.next();
    out << ")";

    return out;
}

inline std::ostream& operator << (std::ostream &out, const BaseFsm::Step &step)
{

    out << step.transition();
    out << (step.is_executed() ? "!" : "?") << " ";
    out << step.message();
    return out;
}

}}} // namespace cpptesk::hw::fsm

