/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHORS:  Alexander Kamkin (kamkin@ispras.ru)
//           Mikhail Chupilko (chupilko@ispras.ru)
// SYNOPSIS: Macros to define mediators (adapters)
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/errors.hpp>
#include <hw/model.hpp>
#include <hw/utils/shared_ptr.hpp>

#include <sstream>

////////////////////////////////////////////////////////////////////////////////
// Declaration of Mediator (Adapter) Classes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_MEDIA
 * Declares a mediator (adapter) of a hardware model.
 */
#define CPPTESK_MEDIA(UserMedia, UserModel) \
    struct UserMedia: public UserModel

/** @def CPPTESK_SIGNAL_MEDIA
 * Declares a signal mediator (adapter) of a hardware model.
 */
#define CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs) \
    struct UserMedia: public UserModel, public ::cpptesk::hw::IOSignals<Inputs, Outputs>

/** @def CPPTESK_ADAPTER
 * Alias of CPPTESK_MEDIA
 */
#define CPPTESK_ADAPTER(UserMedia, UserModel) CPPTESK_MEDIA(UserMedia, UserModel)

/** @def CPPTESK_SIGNAL_ADAPTER
 * Alias of CPPTESK_SIGNAL_MEDIA
 */
#define CPPTESK_SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs) CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)

#ifdef HW_SHORT_NAMES
    #define MEDIA(UserMedia, UserModel)                           CPPTESK_MEDIA(UserMedia, UserModel)
    #define ADAPTER(UserMedia, UserModel)                         CPPTESK_ADAPTER(UserMedia, UserModel)
    #define SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)   CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)
    #define SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs) CPPTESK_SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs)
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// Reaction Arbitration
///////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_DETECT_REACTION_LABEL $CPPTESK_WAIT_REACTION_REACTION

#define __CPPTESK_CHECK_DETECT_REACTION() \
    do { \
        if(false) goto __CPPTESK_DETECT_REACTION_LABEL; \
    } while(false)

#define __CPPTESK_DETECT_REACTION_OF(UnitClass, unit, cond) \
    do { \
        if(_process.listener) { \
            CPPTESK_WAIT_OF(unit, cond); \
            if(_iface.isModelBased()) { \
                CPPTESK_CALL_PROCESS_OF_EX(reinterpret_cast<UnitClass::model_t&>(unit), \
                UnitClass::singleListener, _iface, _message, _START_MODE_OF_LISTENER); \
            } \
        } else { \
            /* Request the interface arbiter */ \
            _iface.appendRequest(_process, _message); \
            /* Only one DETECT_REACTION() (WAIT_REACTION()) is allowed within one deserializer */ \
            __CPPTESK_DETECT_REACTION_LABEL: \
            _message = *__cpptesk_original_reference_model_message; \
            \
            if(getReactionTimeout(_iface) == CPPTESK_INFINITY_TIME) \
                CPPTESK_WAIT_OF(unit, (cond) && _iface.can_adapter_start(_process, _message) && \
                            _iface.getRemoveReactionTimemark() < (cpptesk::hw::cycle_t)(unit).time()); \
            else \
                CPPTESK_WAIT_TIMEOUT_OF(unit, (cond) && _iface.can_adapter_start(_process, _message) && \
                            _iface.getRemoveReactionTimemark() < (cpptesk::hw::cycle_t)(unit).time(), getReactionTimeout(_iface)); \
            \
            if(_iface.some_unexpected_reaction_matches(__start_time, getReactionTimeout(_iface), \
                        getUnexpectedReactionTimeout(_iface), _message)) { \
                _process.timeout = false; \
                _iface.takeOldReaction(); \
                _iface.removeRequest(_process, _message); \
                goto __CPPTESK_NEXT_REACTION_STOP_LABEL; \
            } \
            \
            if((unit).time() >= (getReactionTimeout(_iface) + __start_time) && \
               getReactionTimeout(_iface) != CPPTESK_INFINITY_TIME) { \
                _process.timeout = true; \
                onTimeoutViolation(_process); \
            } \
            /* Missing reaction */ \
            if(_process.timeout) { \
                if(!_iface.remove_spec_reaction_after_timeout(_message)) \
		    _process.timeout = false; \
                goto __CPPTESK_NEXT_REACTION_LABEL; \
            } \
        } \
    } while(false)

/**
 * @def CPPTESK_DETECT_REACTION_OF
 * Detects a reaction by using a single-cycle condition but do not take it.
 * It is used when a detector is multi-cycle. (multi-model case)
 */
#define CPPTESK_DETECT_REACTION_OF(UnitClass, unit, cond) \
    cpptesk::hw::Process &_process = CPPTESK_GET_PROCESS(); \
    cpptesk::hw::Interface &_iface = CPPTESK_GET_IFACE(); \
    cpptesk::hw::Message &_message = CPPTESK_GET_MESSAGE(); \
    cpptesk::hw::SharedPtr<cpptesk::hw::Message> __cpptesk_original_reference_model_message (_message.clone()); \
    bool __user_used_expect_reaction = false; \
    cpptesk::hw::cycle_t __start_time = (unit).time(); \
    __CPPTESK_DETECT_REACTION_OF(UnitClass, unit, cond)

/**
 * @def CPPTESK_DETECT_REACTION
 * Detects a reaction by using a single-cycle condition but do not take it.
 * It is used when a detector is multi-cycle. (single-model case)
 */
#define CPPTESK_DETECT_REACTION(cond) \
    CPPTESK_DETECT_REACTION_OF(model_t, *this, cond)

#define __CPPTESK_TAKE_REACTION() \
    do { \
        /*--------------------------------------------------------------------*/ \
        /* Unexpected Reaction Listener                                       */ \
        /*--------------------------------------------------------------------*/ \
        if(_process.listener) { \
            /* Do nothing */ \
        } \
        /*--------------------------------------------------------------------*/ \
        /* Reaction Deserializer                                              */ \
        /*--------------------------------------------------------------------*/ \
        else { \
            /* Paranoid check */ \
            if(_process.timeout) { \
                CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_WARN, \
                    "Time limit in WAIT_TIMEOUT() has been reached"); \
            } \
            /* Error if a reaction has been taken by someone else in case of modelBased arbiters */ \
            if(_iface.isModelBased()) { \
                if(!_iface.takeReaction(time())) { \
                    CPPTESK_ASSERTION(false, __CPPTESK_COMMON_INFO() << ": " << \
                    "Conflict of model reactions (use SET_ARBITER() to set an arbiter)"); \
                } \
                _iface.removeRequest(_process, _message, time()); \
            } \
        } \
    } while(false)

/**
 * @def CPPTESK_TAKE_REACTION
 * Takes a reaction (says to the interface that the reaction is taken).
 */
#define CPPTESK_TAKE_REACTION() \
    do { \
        ::cpptesk::hw::Process &_process = CPPTESK_GET_PROCESS(); \
        ::cpptesk::hw::Interface &_iface = CPPTESK_GET_IFACE(); \
        ::cpptesk::hw::Message &_message = CPPTESK_GET_MESSAGE(); \
        __CPPTESK_TAKE_REACTION(); \
    } while(false)

/** @def CPPTESK_WAIT_REACTION_OF
 * Detects a reaction by using a single-cycle condition and takes it (multi-model case).
 */
#define CPPTESK_WAIT_REACTION_OF(UnitClass, unit, cond) \
    cpptesk::hw::Process &_process = CPPTESK_GET_PROCESS(); \
    cpptesk::hw::Interface &_iface = CPPTESK_GET_IFACE(); \
    cpptesk::hw::Message &_message = CPPTESK_GET_MESSAGE(); \
    cpptesk::hw::SharedPtr<cpptesk::hw::Message> __cpptesk_original_reference_model_message (_message.clone()); \
    bool __user_used_expect_reaction = false; \
    cpptesk::hw::cycle_t __start_time = (unit).time(); \
    if(false) goto __CPPTESK_WAIT_REACTION_LABEL; \
    __CPPTESK_WAIT_REACTION_LABEL: \
    __user_used_expect_reaction = false; \
    __start_time = (unit).time(); \
    \
    if(!_process.listener) { \
        _iface.add_reaction_to_spec_reactions(_message, (unit).time()); \
    } \
    \
    do { \
        __CPPTESK_DETECT_REACTION_OF(UnitClass, unit, cond); \
        /* Reaction is taken only if time limit is not reached */ \
        if(!_process.timeout) { \
            __CPPTESK_TAKE_REACTION(); \
        } \
    } while(false)

/** @def CPPTESK_WAIT_REACTION
 * Detects a reaction by using a single-cycle condition and takes it (single-model case).
 */
#define CPPTESK_WAIT_REACTION(cond) \
    CPPTESK_WAIT_REACTION_OF(model_t, *this, cond)

/** @def CPPTESK_EXPECT_REACTION
 * Enables next reactions to come and be deserialized by the adapter.
 */
#define CPPTESK_EXPECT_REACTION() \
    do { \
        if(CPPTESK_GET_PROCESS().listener) { \
            CPPTESK_ASSERTION(!__user_used_expect_reaction, "Macro CPPTESK_EXPECT_REACTION() can be used only once in one deserializer."); \
            CPPTESK_GET_IFACE().setDetectionFinished(true); \
            __user_used_expect_reaction = true; \
        } \
    } while(false)

/** @def __CPPTESK_EXPECT_REACTION
 * Enables next reactions to come and be deserialized by the adapter (to be used in NEXT_REACTION macro).
 */
#define __CPPTESK_EXPECT_REACTION() \
    do { \
        if(CPPTESK_GET_PROCESS().listener && !__user_used_expect_reaction) \
            CPPTESK_GET_IFACE().setDetectionFinished(true); \
        __user_used_expect_reaction = false; \
    } while(false)

/** @def CPPTESK_NEXT_REACTION_OF
 * Finishes processing of a reaction and switches to the next one. (multi-model case)
 */
#define CPPTESK_NEXT_REACTION_OF(UnitClass, unit) \
    do { \
        __CPPTESK_NEXT_REACTION_LABEL: \
        \
        /*--------------------------------------------------------------------*/ \
        /* Unexpected Reaction Listener                                       */ \
        /*--------------------------------------------------------------------*/ \
        if(_process.listener) \
        { \
            if(!_iface.isModelBased()) \
            { \
                CPPTESK_CALL_PROCESS_OF_EX(reinterpret_cast<UnitClass::model_t&>(unit), \
                UnitClass::singleListener, _iface, _message, _START_MODE_OF_LISTENER); \
            } \
            \
            __CPPTESK_EXPECT_REACTION(); \
            \
            if(getUnexpectedReactionTimeout(_iface) != 0) \
                CPPTESK_DELAY_OF(unit, getUnexpectedReactionTimeout(_iface) + 1); \
            else \
                CPPTESK_CYCLE_OF(unit); \
        } \
        /*--------------------------------------------------------------------*/ \
        /* Reaction Deserializer                                              */ \
        /*--------------------------------------------------------------------*/ \
        \
        else { \
            /* REMEMBER: the following section is started only for non-listeners! */ \
            \
            /* Check that DETECT_REACTION() (WAIT_REACTION()) exists */ \
            __CPPTESK_CHECK_DETECT_REACTION(); \
            \
            /* If the adapter uses the compare arbiter and the timeout is not still violated, */ \
            /* one should try to find appropriate message in the arbiter msg-list. If the message is found, Adapter tries to take it. */ \
            /* If the one didn't succeed in taking, it means that some other adapter had already taken the message */ \
            /* and this adapter should start checking again */ \
            \
            if(!_iface.isModelBased() && !_process.timeout) \
            { \
                if(!_iface.chooseRequest(_process, _message)) \
                { \
                    CPPTESK_CYCLE_OF(unit); \
                    goto __CPPTESK_DETECT_REACTION_LABEL; \
                } \
                else if(!_iface.takeReaction((unit).time())) \
                { \
                    CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_WARN, __CPPTESK_COMMON_INFO() << ": " << \
                        "The compare arbiter cannot make a decision (two model messages are the same)"); \
                    CPPTESK_CYCLE_OF(unit); \
                    goto __CPPTESK_DETECT_REACTION_LABEL; \
                } \
                \
                _iface.removeRequest(_process, _message, (unit).time()); \
            } \
        } \
        __CPPTESK_NEXT_REACTION_STOP_LABEL: \
        while(false) {}; \
    } while(false)

/** @def CPPTESK_NEXT_REACTION
 * Finishes processing of a reaction and switches to the next one. (single-model case)
 */
#define CPPTESK_NEXT_REACTION() \
    CPPTESK_NEXT_REACTION_OF(model_t, *this)

#ifdef HW_SHORT_NAMES
    #define DETECT_REACTION(cond)                     CPPTESK_DETECT_REACTION(cond)
    #define TAKE_REACTION()                           CPPTESK_TAKE_REACTION()
    #define WAIT_REACTION(cond)                       CPPTESK_WAIT_REACTION(cond)
    #define NEXT_REACTION()                           CPPTESK_NEXT_REACTION()
    #define EXPECT_REACTION()                         CPPTESK_EXPECT_REACTION()
    #define DETECT_REACTION_OF(UnitClass, unit, cond) CPPTESK_DETECT_REACTION(UnitClass, unit, cond)
    #define WAIT_REACTION_OF(UnitClass, unit, cond)   CPPTESK_WAIT_REACTION_OF(UnitClass, unit, cond)
    #define NEXT_REACTION_OF(UnitClass, unit)         CPPTESK_NEXT_REACTION_OF(UnitClass, unit)
#endif // HW_SHORT_NAMES

namespace cpptesk {
namespace hw {

}} // namespace cpptesk::hw
