/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of an operation execution process
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/event.hpp>
#include <hw/priority.hpp>
#include <hw/time.hpp>

#include <stdint.h>

#include <iostream>
#include <sstream>
#include <list>
#include <string>

namespace cpptesk {
namespace hw {

class Interface;
class Message;
class ProcessTable;

#define CPPTESK_INCORRECT_PID (-1)

#ifdef HW_SHORT_NAMES
    #define INCORRECT_PID CPPTESK_INCORRECT_PID
#endif // HW_SHORT_NAMES

/** @typedef pid_t
 * Process identifier.
 */
typedef int pid_t;

/** @typedef status_t
 * Process status.
 */
typedef enum status_t
{
    /// Process is still running.
    EXECUTING = -1,
    /// Process execution has been successfully completed.
    COMPLETED =  0,
    /// Precondition violation.
    UNPREDICTED =  1,
    /// Postcondition violation.
    VIOLATION =  2
} status_t;

/** @typedef point_t
 * Synchronization point type.
 */
typedef enum point_t
{
    /// Initial synchronization point (process is not executed).
    START = 0,
    /// Synchronization caused by CYCLE() or DELAY().
    CYCLE = 1,
    /// Synchronization caused by GUARD(condition) or other kind of wait condition.
    GUARD = 2,
    /// Other types of points.
    OTHER = -1
} point_t;

class Process
{
    friend class ContextSwitcher;
    friend class ProcessScheduler;
    friend class ProcessTable;

public:
    /// Process attributes
    struct Attributes
    {
        Attributes():
            type(0), priority(NORMAL_PRIORITY), clone_message(false)
        {}

        Attributes(unsigned type, unsigned priority, bool clone_message):
            type(type), priority(priority), clone_message(clone_message)
        {}

        unsigned type;
        unsigned priority;
        bool clone_message;
    };

    /// Descriptor of a point in code in which a process is called.
    struct CallDescriptor
    {
        CallDescriptor() { reset(); }

        CallDescriptor(const std::string &name, const std::string &file = "<unknown>", int line = -1):
            name(name), file(file), line(line) {}

        void reset()
        {
            name = "<unknown>";
            file = "<unknown>";
            line = -1;
        }

        /// Name of a process function.
        std::string name;
        /// File name where a process is called.
        std::string file;
        /// Line in the file where a process is called.
        int line;
    };

    /// Descriptor of a synchronization point within a process.
    struct PointDescriptor
    {
        PointDescriptor(point_t type = START, const std::string &name = "START()", cycle_t now = CPPTESK_INFINITY_TIME,
	    const std::string &file = "<unknown>", int line = -1):
            type(type), name(name), file(file), line(line), time(now) {}

        /// Point type.
        point_t type;
        /// Point name.
        std::string name;
        /// File name where a point is defined.
        std::string file;
        /// Line in the file where a point is defined.
        int line;
        /// Point timemark.
        cycle_t time;
    };

    /// Call stack.
    typedef std::list<CallDescriptor> call_stack_t;

private:
    /// 
    Process(pid_t pid, ProcessTable &process_table);
    Process(pid_t pid, cycle_t now, Interface &iface, Message &message,
        Attributes attrs, ProcessTable &process_table);
    ~Process();

public:
    pid_t get_pid() const { return _pid; }

    unsigned get_priority() const { return _attrs.priority; }
    void set_priority(unsigned v);

    status_t get_status() const { return _status; }
    void set_status(status_t v) { _status = v; }

    /// Returns the process start time.
    cycle_t start_time() const { return _start_time; }

    const PointDescriptor& get_point() const { return _point; }
    void set_point(const PointDescriptor &point) { _point = point; }

    /// Prepares a call descriptor of a process.
    void prepare_call(const CallDescriptor &call)
    {
        if(call_stack.empty())
            { parallel_call = call; }

        sequential_call = call;
    }

    /// Returns size of the call stack.
    size_t call_stack_size() const { return call_stack.size();  }

    /// Returns size of the call stack.
    const CallDescriptor& call_stack_top()    const { return call_stack.back();   }
    const CallDescriptor& call_stack_bottom() const { return call_stack.front();  }

    /// Adds a call descriptor into the stack.
    void start_call()
        { call_stack.push_back(sequential_call); sequential_call.reset(); }

    /// Removes a call descriptor from the stack. Returns true iff it's the last item.
    void stop_call()
        { call_stack.pop_back(); }

    /// Descriptor of the process call (bottom of the call stack).
    CallDescriptor parallel_call;
    /// Descriptor of a currently called sequential process within the process.
    CallDescriptor sequential_call;

    /// Stack of sequential calls within the process (contains this process's call in the bottom).
    call_stack_t call_stack;

    /// Shortcuts to the process table methods.
    void resume();
    void suspend();
    void suspend(cycle_t next_time_slot);
    void schedule_process();
    void schedule_process(cycle_t time_slot);

    /// Cause-effect relation
    void set_cause(const event_ptr_t &event);
    void add_effect(event_ptr_t &event);
    void caused_by(const Process &process);

///////////////////////////////////////////////////////////////////////////////
// For internal use only (these members are likely to be depricated)
///////////////////////////////////////////////////////////////////////////////

    /// Timeout flag.
    bool timeout;
    /// Input interface.
    Interface *input_iface;
    /// Copy of the input message.
    Message *input_message;
    /// The original input message.
    Message *original_message;
    /// Unexpected reaction listener.
    bool listener;

private:
    /// Returns the current cycle of execution (if current_time() == schedule_time())
    /// or the recently processed cycle (if current_time() != schedule_time()).
    cycle_t current_time() const { return _current_time; }
    void current_time(cycle_t v) { _current_time = v; }

    /// Returns/sets the schedule time (next execution time slot).
    cycle_t schedule_time() const { return _schedule_time; }
    void schedule_time(cycle_t v) { _schedule_time = v; }

    /// Returns true iff the execution cycle is completed (current_time() != schedule_time()).
    bool is_cycle_completed() const { return _current_time != _schedule_time; }

    /// Returns true iff the process is fully completed.
    bool is_completed() const { return _status != EXECUTING; }

    /// Gets/sets the process context.
    template <typename T>  T* get_context() { return (T*)_context; }
    void set_context(void *context) { _context = context; }

private:
    /// Process identifier.
    pid_t _pid;
    /// Status of the process execution.
    status_t _status;
    /// Differen timemarks/
    cycle_t _start_time;
    cycle_t _current_time;
    cycle_t _schedule_time;
    /// Current synchronization point.
    PointDescriptor _point;
    /// Pointer to the process context (context switcher-dependent data).
    void* _context;
    /// Process attributes.
    Attributes _attrs;
    /// Reference to the process table.
    ProcessTable &_process_table;

    /// Cause-effect relation.
    event_ptr_t  _cause;
    event_list_t _effects;
};

inline std::ostream& operator << (std::ostream& out, const Process::CallDescriptor &call)
{
    // Parameters are unsupported
    out << call.name << "()";

    if(call.line >= 0)
        { out << " at " << call.file << ":" << std::dec << call.line << std::hex; }

    return out;
}

inline std::ostream& operator << (std::ostream& out, const Process::call_stack_t &call_stack)
{
    Process::call_stack_t::const_reverse_iterator i;

    // IMPORTANT: Call stack can be empty before START()
    if(call_stack.empty()) { return out; }

    out << " in " << *(i = call_stack.rbegin());

    for(++i; i != call_stack.rend(); i++)
        { out << std::endl << " in " << *i; }

    return out;
}

inline std::ostream& operator << (std::ostream& out, const Process &process)
{
    // IMPORTANT: Call stack can be empty before START()
    const Process::CallDescriptor &call = (process.call_stack_size() > 0 ? process.call_stack_top() : process.parallel_call);

    return out << call.name << "(pid=" << process.get_pid() << ")";
}

}} // namespace cpptesk::hw

