/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Sequence recognizer
///////////////////////////////////////////////////////////////////////////////

#include <hw/assertion.hpp>
#include <hw/sequence/recognizer.hpp>

namespace cpptesk {
namespace hw {
namespace sequence {

////////////////////////////////////////////////////////////////////////////////
// Sequence Recognizer
////////////////////////////////////////////////////////////////////////////////

unsigned SequenceRecognizer::cycle()
{
    unsigned new_match = 0;
    SequenceRecognizer *recognizer = 0;

    // If the recognizer has finished, processing is not performed
    if(isFinished()) { return 0; }

    unsigned old_level = CPPTESK_SET_DEBUG_INDENT(getLevel());
    CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL, "SequenceRecognizer::cycle: time=" << time);

    if(delay.min > 0)
        { goto tick_recognizer; }

    if(getTime() > time_constraints.max)
        { goto stop_recognizer; }

    // If time limit is not reached, create one more delay alternative
    if(!timeout)
    {
CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_TOOL, "-----------------add new----------------");
        switch(sequence.getType())
        {
        case Sequence::EVENT:  { recognizer = new EventRecognizer (sequence,    this); break; }
        case Sequence::GUARD:  { recognizer = new GuardRecognizer (sequence,    this); break; }
        case Sequence::SEQ:    { recognizer = new SeqRecognizer   (sequence, i, this); break; }
        case Sequence::AND:    { recognizer = new AndRecognizer   (sequence,    this); break; }
        case Sequence::OR:     { recognizer = new OrRecognizer    (sequence,    this); break; }
        case Sequence::REPEAT: { recognizer = new RepeatRecognizer(sequence,    this); break; }
        }
        CPPTESK_ASSERTION(recognizer, "Cannot find a recognizer: unknown sequence type");
        alt.push_back(recognizer);
    }

CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_TOOL, "------------------for cycle-------------");
    for(std::list<SequenceRecognizer*>::iterator i = alt.begin(); i != alt.end();)
    {
CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_TOOL, "-----------------------onCycle---------------");
        new_match += (*i)->onCycle();

	if((*i)->isFinished())
	    { delete (*i); i = alt.erase(i); }
        else
            { i++; }
    }

    if(delay.max == 0)
    {
        if(alt.empty())
            { goto stop_recognizer; }
	else
            { timeout = true; }
    }

tick_recognizer:
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "SequenceRecognizer::cycle: delay=[" << delay.min << "," << delay.max << "]");
        CPPTESK_SET_DEBUG_INDENT(old_level);

        tick();
        return match(new_match);
    }

stop_recognizer:
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "SequenceRecognizer::cycle: stop");
        CPPTESK_SET_DEBUG_INDENT(old_level);

        stop();
        return match(new_match);
    }
}

unsigned SequenceRecognizer::onCycle()
{
    return 0;
}

void SequenceRecognizer::reset()
{
    start();

    delay = initial_delay;
    count = time = 0;

    timeout = false;

    if(time_constraints.min > delay.min)
        { delay.min = time_constraints.min; }
    if(time_constraints.max < delay.max)
        { delay.max = time_constraints.max; }
    if(delay.min > delay.max)
        { stop(); }

    for(std::list<SequenceRecognizer*>::iterator i = alt.begin(); i != alt.end(); i++)
        { delete *i; }

    alt.clear();
}

size_t SequenceRecognizer::size() const
{
    // Take into account this recognizer
    size_t count = 1;

    for(std::list<SequenceRecognizer*>::const_iterator i = alt.begin(); i != alt.end(); i++)
        { count += (*i)->size(); }

    return count;
}

////////////////////////////////////////////////////////////////////////////////
// EventRecognizer
////////////////////////////////////////////////////////////////////////////////

unsigned EventRecognizer::onCycle()
{
    unsigned old_indent = CPPTESK_SET_DEBUG_INDENT(getLevel());
    CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL, "EventRecognizer::onCycle");

    unsigned new_match = match(sequence.getEvent().match() ? 1 : 0);

    stop();

    CPPTESK_SET_DEBUG_INDENT(old_indent);
    return new_match;
}

////////////////////////////////////////////////////////////////////////////////
// GuardRecognizer
////////////////////////////////////////////////////////////////////////////////

unsigned GuardRecognizer::onCycle()
{
    unsigned old_indent = CPPTESK_SET_DEBUG_INDENT(getLevel());
    CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL, "GuardRecognizer::onCycle");

    unsigned new_match = match(sequence.getEvent().match() ? 1 : 0);

    if(new_match > 0) { stop(); }

    CPPTESK_SET_DEBUG_INDENT(old_indent);
    return new_match;
}

////////////////////////////////////////////////////////////////////////////////
// SeqRecognizer
////////////////////////////////////////////////////////////////////////////////

unsigned SeqRecognizer::onCycle()
{
    unsigned old_indent = CPPTESK_SET_DEBUG_INDENT(getLevel());
    CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL, "SeqRecognizer::onCycle");

    // Empty sequence is trivially processed
    if(!main) { stop(); return match(1); }

    // Match count for the overall sequence
    unsigned new_match = 0;
    // Process the current item of the sequence
    unsigned main_inner_match = main->cycle();

    // If the current item has been successfully matched, process all
    // zero-cycle items and one cycle of the first multi-cycle item:
    // 
    //  Event1 #1 | Event2 #1 | ... #1 | EventN
    //  cycle()   | cycle()   | ...    | cycle()
    //
    bool delta_delay_cycle = main->isFinished();
    while(main_inner_match > 0 && main->isFinished() && delta_delay_cycle)
    {
        // Break if the end of the sequence is reached
        if(++i == sequence.end())
	    { break; }

        // Go to the next item (if possible)
        delete main; main = new SequenceRecognizer(*i, this);

	delta_delay_cycle &= (main->getInitialDelay().isZero());
        main_inner_match = main->cycle();
    }

// TODO: ASSERT delta_delay_cycle == main->isFinished() ?
// TODO: Zero delay alternative: Event1 << #0:1 << Event2

    // The current item has been successfully matched
    if(main_inner_match > 0)
    {
        // The recognizer of the current item is still active (multiple matching is possible)
        if(main->isWorking())
        {
	    // Go to the next item of the sequence
	    i++;
            // If it is available, recognition bifurcates
            if(i != sequence.end())
// TODO: Optimization: if the recognizer for that i (R[i]) has been already added, then we can simply
// TODO: reinitialize the delay range (R[i]->delay = initial_delay) without adding a new alternative.
// TODO: The optimization can lead to the incorrent match counting, but it reduces the memory usage.
                { CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_TOOL, "alternative"); alt.push_back(new SeqRecognizer(sequence, i, this)); }
            // If the current item is the last one, a new match is found
            else
                { new_match = main_inner_match; }
            // Come back to the current item (it is still under processing)
            i--;
        }
        // The recognizer of the current item has been successfully finished
        else
        {
	/*
            // Go to the next item of the sequence
            i++;
            // If it is available, change the current item
            if(i != sequence.end())
            {
                CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_TOOL, "next");

                delete main; main = new SequenceRecognizer(*i, this);
            }
            // If the current item is the last one, a new match is found
            else
	 */
                { new_match = main_inner_match; }
        }
    }

CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_TOOL, "-----------------------------------alternatives--------------------------" << alt.size());

    // Check the results of the alternative recognizers
    for(std::list<SequenceRecognizer*>::iterator i = alt.begin(); i != alt.end();)
    {
        // If at least one of the recognizers is successful, the overall recognizer is successful as well
CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_TOOL, "----------------------------------cycle--------------------------"); 
        new_match += (*i)->cycle();

        // Erase the recognizer if it is finished
        if((*i)->isFinished())
            { CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_TOOL, "-----delete-----"); delete (*i); i = alt.erase(i); }
        else
            { i++; CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_TOOL, "------i++------"); }
    }

    // The recognizer finishes when the main recognizer and all alternative ones are completed
    if(main->isFinished() && alt.empty())
        { stop(); }

    CPPTESK_SET_DEBUG_INDENT(old_indent);
    return match(new_match);
}

////////////////////////////////////////////////////////////////////////////////
// AndRecognizer
////////////////////////////////////////////////////////////////////////////////

unsigned AndRecognizer::onCycle()
{
    unsigned old_indent = CPPTESK_SET_DEBUG_INDENT(getLevel());
    CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL, "AndRecognizer::onCycle");

    unsigned new_lhs_match = lhs->cycle();
    unsigned new_rhs_match = rhs->cycle();

    if(lhs->isFinished() && rhs->isFinished()) { stop(); }

    CPPTESK_SET_DEBUG_INDENT(old_indent);
    return match((rhs->getMatchCount() > 0 ? new_lhs_match : 0)
               + (lhs->getMatchCount() > 0 ? new_rhs_match : 0));
}

/////////////////////////////////////////////////////////////////////////////////
// OrRecognizer
/////////////////////////////////////////////////////////////////////////////////

unsigned OrRecognizer::onCycle()
{
    unsigned old_indent = CPPTESK_SET_DEBUG_INDENT(getLevel());
    CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL, "OrRecognizer::onCycle");

    unsigned new_lhs_match = lhs->cycle();
    unsigned new_rhs_match = rhs->cycle();

    if(lhs->isFinished() && rhs->isFinished()) { stop(); }

    CPPTESK_SET_DEBUG_INDENT(old_indent);
    return match(new_lhs_match + new_rhs_match);
}

////////////////////////////////////////////////////////////////////////////////
// RepeatRecognizer
////////////////////////////////////////////////////////////////////////////////

unsigned RepeatRecognizer::onCycle()
{
    unsigned old_indent = CPPTESK_SET_DEBUG_INDENT(getLevel());
    CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL, "AndRecognizer::onCycle");

    unsigned new_match = 0;
    unsigned new_inner_match = main->cycle();

    if(new_inner_match)
    {
        if(main->isWorking())
        {
            if(repeat.max > 1)
            {
                Repeat dec_repeat((repeat.min > 0 ? repeat.min - 1 : 0),
                                  (repeat.max > 0 ? repeat.max - 1 : 0));

                alt.push_back(new RepeatRecognizer(sequence, dec_repeat, this));
            }
        }
        else
        {
            if(repeat.min > 0) { repeat.min--; }
            if(repeat.max > 0) { repeat.max--; }

            if(repeat.min == 0)
                { new_match = 1; }

            if(repeat.max > 0)
                { delete main; main = new SequenceRecognizer(sequence.getChild(), this); }
        }
    }

    for(std::list<SequenceRecognizer*>::iterator i = alt.begin(); i != alt.end(); i++)
    {
        new_match += (*i)->cycle();

        if((*i)->isFinished())
            { delete *i; alt.erase(i--); }
    }

    if(main->isFinished() && alt.empty())
        { stop(); }

    CPPTESK_SET_DEBUG_INDENT(old_indent);
    return match(new_match);
}

}}} // namespace cpptesk::hw::sequence

