/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Generalized model state 
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>
#include <hw/unused.hpp>

#include <iterator>
#include <ostream>
#include <set>
#include <string>
#include <vector>

namespace cpptesk {
namespace hw {
namespace state {

///////////////////////////////////////////////////////////////////////////////

class SingleState
{
public:
    SingleState() {}
   ~SingleState() {}
};

inline std::ostream& operator << (std::ostream &out, const SingleState &state)
{
    CPPTESK_UNUSED_ARG(state);
    return out << "single[]";
}

///////////////////////////////////////////////////////////////////////////////

template<typename T>
class AtomicState
{
public:
    AtomicState(const T &state): _state(state) {}
   ~AtomicState() {}

    const T& state() const { return _state; }

private:
    T _state;
};

template<typename T>
inline std::ostream& operator << (std::ostream &out, const AtomicState<T> &state)
{
    return out << "atomic[" << state.state() << "]";
}

///////////////////////////////////////////////////////////////////////////////

template<typename T, class T_collection>
class CompositeState
{
public:
    typedef typename T_collection::iterator iterator;
    typedef typename T_collection::const_iterator const_iterator;

    template<class InputIterator>
    CompositeState(InputIterator first, InputIterator last)
    {
        _state.insert(first, last);
    }

    template<class InputIterator>
    CompositeState(InputIterator first, InputIterator last, bool /* dummy */)
    {
        _state.insert(_state.begin(), first, last);
    }

    iterator begin() { return _state.begin(); }
    const_iterator begin() const { return _state.begin(); }

    iterator end() { return _state.end(); }
    const_iterator end() const { return _state.end(); }

protected:
    T_collection _state;
};

template<typename T, class T_collection>
inline std::ostream& operator << (std::ostream &out, const CompositeState< T, T_collection > &state)
{
    out << "[";

    for(typename T_collection::const_iterator i = state.begin(); i != state.end(); i++)
        { out << *i; }

    return out << "]";
}

///////////////////////////////////////////////////////////////////////////////

template<typename T>
class VectorState: public CompositeState< T, std::vector<T> >
{
public:
    template<class InputIterator>
    VectorState(InputIterator first, InputIterator last): CompositeState< T, std::vector<T> >(first, last, true) {}
};

template<typename T>
inline std::ostream& operator << (std::ostream &out, const VectorState<T> &state)
{
    return out << "vector" << CompositeState< T, std::vector<T> >(state);
}

///////////////////////////////////////////////////////////////////////////////

template<typename T>
class MultisetState: public CompositeState< T, std::multiset<T> >
{
public:
   template<class InputIterator>
   MultisetState(InputIterator first, InputIterator last): CompositeState< T, std::multiset<T> >(first, last) {}
};

template<typename T>
inline std::ostream& operator << (std::ostream &out, const MultisetState<T> &state)
{
    return out << "multiset" << CompositeState< T, std::multiset<T> >(state);
}

///////////////////////////////////////////////////////////////////////////////

template<typename T>
class SetState: public CompositeState< T, std::set<T> >
{
public:
   template<class InputIterator>
   SetState(InputIterator first, InputIterator last): CompositeState< T, std::set<T> >(first, last) {}
};

template<typename T>
inline std::ostream& operator << (std::ostream &out, const SetState<T> &state)
{
    return out << "set" << CompositeState< T, std::set<T> >(state);
}

}}} // namespace cpptesk::hw::state

