/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <netfsm2.h>
#include <atl/arraylist.h>
#include <atl/list.h>
#include <atl/string.h>
#include <sgl/graph.h>
#include <sgl/utils.h>
#include <utils/assertion.h>

// Base FSM engines
#include <netndfsm.h>
#include <netfsm.h>
#include <netpfsm.h>

//------------------------------------------------------------------------------
// FSM2 Macros
//------------------------------------------------------------------------------

#define FSM2_BASE_ENGINE     ENGINE_FSM
#define FSM2_FSM_NUMBER      -1
#define FSM2_STATE_DELIMITER ';'
#define FSM2_RANDOMIZE       0
#define FSM2_DATA_ALLOC_BY   16

//------------------------------------------------------------------------------
// FSM2 Data
//------------------------------------------------------------------------------

typedef struct FSM2_Data
{
    /// The reference to the external graph information.
    TE_GraphInfo *info;

    /// The test engine.
    TestEngine *engine;
    
    /// The matrix of graphs (engines data).
    ArrayList data;

    /// The number of component FSMs.
    int number;
    /// The state delimiter.
    char delimiter;

    /// Randomize choice of an untraversed edge.
    int randomize;
    
    /// The recently used engine.
    int i;
    int j;
} FSM2_Data;

static FSM2_Data* alloc_Data(void)
{
    FSM2_Data *data = (FSM2_Data*)malloc(sizeof(FSM2_Data));
    assertion(data != NULL, "alloc_Data: data == NULL");
    
    return data;
}

static void dealloc_Data(FSM2_Data *data)
{
    if(data != NULL) { free(data); }
}

static void init_Data(FSM2_Data *data, TE_GraphInfo *info)
{
    data->info = info;

    init_ArrayList(&data->data, sizeof(void*), FSM2_DATA_ALLOC_BY, NULL);

    data->engine    = FSM2_BASE_ENGINE;
    data->number    = FSM2_FSM_NUMBER;
    data->delimiter = FSM2_STATE_DELIMITER;
    data->randomize = FSM2_RANDOMIZE;
    
    data->i = 0;
    data->j = 1;
}

static void destroy_Data(FSM2_Data *data)
{
    destroy_ArrayList(&data->data);
}

//------------------------------------------------------------------------------
// FSM2 Interface
//------------------------------------------------------------------------------

static void processParams_FSM2(TE_Graph *graph, int argc, char **argv)
{
    int i;

    FSM2_Data *data = (FSM2_Data*)graph->data;

    assertion(argc >= 0,                 "processParams_FSM2: argc < 0");
    assertion(argv != NULL || argc == 0, "processParams_FSM2: argv == NULL && argc > 0");
    
    for(i = 0; i < argc; i++)
    {
        if(!strcmp("--number", argv[i]))
        {
            assertion(++i < argc, "processParams_FSM2: --number should be followed by an integer number");
            data->number = atoi(argv[i]);
            assertion(data->number >= 2, "processParams_FSM2: --incorrect value of --number (should be >= 2)");
        }
        else if(!strcmp("--delimiter", argv[i]))
        {
            assertion(++i < argc && strlen(argv[i]) == 1, "processParams_FSM2: --delimiter should be followed by a character");
            data->delimiter = argv[i][0];
        }
        else if(!strcmp("--engine", argv[i]))
        {
            assertion(++i < argc, "processParams_FSM2: --engine should be followed by an engine name (fsm or pfsm)");
            
            if(!strcmp("fsm", argv[i]))
                { data->engine = ENGINE_FSM; }
            else if(!strcmp("pfsm", argv[i]))
                { data->engine = ENGINE_PFSM; }
            else
                { assertion(false, "processParams_FSM2: incorrect value of --engine (should be fsm or pfsm)"); }
        }
        else if(!strcmp("--randomize", argv[i]))
        {
            data->randomize = 1;
        }        
    }

    assertion(data->number >= 2, "processParams_FSM2: use --number to specify the number of components");
}

static void init_FSM2(TE_Graph *graph, int argc, char **argv)
{
    int i, j;
    FSM2_Data *data = alloc_Data();

    init_Data(data, &graph->info);
    graph->data = data;
    processParams_FSM2(graph, argc, argv);

    for(i = 0; i < data->number; i++)
    for(j = 0; j < data->number; j++)
    {
        void** view = (void**)append_ArrayList(&data->data);
        *view = NULL;

        if(j > i)
        {
            graph->data = NULL;
            data->engine->init(graph, argc, argv);
           *view = graph->data;
        }
    }
    
    graph->data = data;
}

static void config_FSM2(TE_Graph *graph, const char *attr_name, const char *attr_value)
{
    int i, j;
    FSM2_Data *data = (FSM2_Data*)graph->data;

    for(i = 0;     i < data->number - 1; i++)
    for(j = i + 1; j < data->number    ; j++)
    {
        void **view = (void**)get_ArrayList(&data->data, i*data->number + j);

        graph->data = *view;
        data->engine->config(graph, attr_name, attr_value);
    }
    
    graph->data = data;
}

static List* parse_NodeId(TE_Graph *graph, TE_NodeId *node_id)
{
    int i, j, length;
    const char *string;

    char *buffer;
    List *nodes;

    FSM2_Data *data = (FSM2_Data*)graph->data;

    if(node_id == NULL)
        { return NULL; }

    string = toCharArray_String(r(node_id->id));
    length = length_String(r(node_id->id));

    if(length == 0)
        { return NULL; }

    // Partial node id.
    if(string[0] == data->delimiter)
        { return NULL; }
        
    nodes  = create_List(&type_String);
    buffer = malloc(length);
    
    for(i = j = 0; j < length; j++)
    {
        if(string[j] == data->delimiter || j == length - 1)
        {
            i = buffer[i] = 0;
            append_List(r(nodes), create_String(buffer));
        }
        else
        {
            buffer[i++] = string[j];
        }
    }

    free(buffer);

    assertion(size_List(r(nodes)) == data->number, "parse_NodeId: incorrect node format");
    return nodes;
}

static TE_NodeId* combine_NodeId(int i, int j, char delimiter, String *lhs, String *rhs)
{
    return create_NodeId(format_String("%c[%d,%d]%c%s%c%s",
        delimiter, i, j,
        delimiter, toCharArray_String(r(lhs)),
        delimiter, toCharArray_String(r(rhs))));
}

static TE_UpdateStatus start_FSM2(TE_Graph *graph, TE_NodeId *node_id)
{
    int i, j;
    FSM2_Data *data = (FSM2_Data*)graph->data;

    TE_UpdateStatus status = TE_OLD_ALL;
    List *nodes = parse_NodeId(graph, node_id);

    for(i = 0;     i < data->number - 1; i++)
    for(j = i + 1; j < data->number;     j++)
    {
        void **view = (void**)get_ArrayList(&data->data, i*data->number + j);

        TE_NodeId *node = combine_NodeId(i, j, data->delimiter,
            get_List(r(nodes), i), get_List(r(nodes), j));
        
        graph->data = *view;
        status |= data->engine->start(graph, node);
        
        destroy(node);
    }
    
    graph->data = data;

    destroy(nodes);
    return status;
}

static TE_UpdateStatus update_FSM2(TE_Graph *graph, TE_UpdateType type,
    TE_ArcId *arc_id, TE_NodeId *new_end_node_id)
{
    int i, j;
    FSM2_Data *data = (FSM2_Data*)graph->data;

    TE_UpdateStatus status = TE_OLD_ALL;

    List *start_nodes   = parse_NodeId(graph, arc_id->start_node_id);
    List *end_nodes     = parse_NodeId(graph, arc_id->end_node_id);
    List *new_end_nodes = parse_NodeId(graph, new_end_node_id);

    for(i = 0;     i < data->number - 1; i++)
    for(j = i + 1; j < data->number;     j++)
    {
        void **view = (void**)get_ArrayList(&data->data, i*data->number + j);
        
        TE_NodeId *start_node = start_nodes != NULL ?
            combine_NodeId(i, j, data->delimiter, get_List(r(start_nodes), i),
                                                  get_List(r(start_nodes), j)) : NULL;

        TE_NodeId *end_node = end_nodes != NULL ?
            combine_NodeId(i, j, data->delimiter, get_List(r(end_nodes), i),
                                                  get_List(r(end_nodes), j)) : NULL;

        TE_NodeId *new_end_node = new_end_nodes != NULL ?
            combine_NodeId(i, j, data->delimiter, get_List(r(new_end_nodes), i),
                                                  get_List(r(new_end_nodes), j)) : NULL;

        TE_ArcId *arc = create_ArcId(start_node, arc_id->stimulus_id, end_node);
            
        graph->data = *view;
        status |= data->engine->update(graph, type, arc, new_end_node);
        
        destroy(arc);
        destroy(new_end_node);
    }

    graph->data = data;
    
    destroy(start_nodes);
    destroy(end_nodes);
    destroy(new_end_nodes);

    return status;
}

static TE_ArcList* getPath_FSM2(TE_Graph *graph, TE_NodeId *node_id)
{
    int i, j, k;
    int error_i, error_j, error = 0;

    FSM2_Data *data = (FSM2_Data*)graph->data;
    
    int start_i = data->randomize ? rand() : 0;
    int start_j = data->randomize ? rand() : 0;
    
    TE_ArcList *path = NULL;
    List *nodes = parse_NodeId(graph, node_id);

    for(i = 0; i < data->number - 1; i++)
    {
        data->i = (start_i + i) % (data->number - 1);
        
        for(j = data->i + 1; j < data->number; j++)
        {
            data->j = (start_j + j) % (data->number - (data->i + 1)) + (data->i + 1);

            void **view = (void**)get_ArrayList(&data->data, data->i*data->number + data->j);

            TE_NodeId *node = combine_NodeId(data->i, data->j, data->delimiter,
                get_List(r(nodes), data->i), get_List(r(nodes), data->j));
            
            graph->data = *view;
            path = data->engine->getPath(graph, node);
            
            destroy(node);
            
            if(path != NULL) { error = 0; goto end_loop; }
            
            if(graph->error != 0)
            {
                error = graph->error;

                error_i = data->i;
                error_j = data->j;
                
                graph->error = 0;
            }
        }
    }

end_loop:    
    
    graph->data  = data;
    graph->error = error;

    if(error)
    {
        data->i = error_i;
        data->j = error_j;
    }
    
    destroy(nodes);
    return path;
}

static TE_ArcList* reachNode_FSM2(TE_Graph *graph, TE_NodeId *node_id, TE_ArcId *arc_id)
{
    int i, j;
    FSM2_Data *data = (FSM2_Data*)graph->data;

    TE_ArcList *path = NULL;

    List *nodes       = parse_NodeId(graph, node_id);
    List *start_nodes = parse_NodeId(graph, arc_id != NULL ? arc_id->start_node_id : NULL);
    List *end_nodes   = parse_NodeId(graph, arc_id != NULL ? arc_id->end_node_id   : NULL);

    for(i = 0;     i < data->number - 1; i++)
    for(j = i + 1; j < data->number;     j++)
    {
        void **view = (void**)get_ArrayList(&data->data, i*data->number + j);

        TE_NodeId *node = nodes != NULL ?
            combine_NodeId(i, j, data->delimiter, get_List(r(nodes), i),
                                                  get_List(r(nodes), j)) : NULL;

        TE_NodeId *start_node = start_nodes != NULL ?
            combine_NodeId(i, j, data->delimiter, get_List(r(start_nodes), i),
                                                  get_List(r(start_nodes), j)) : NULL;

        TE_NodeId *end_node = end_nodes != NULL ?
            combine_NodeId(i, j, data->delimiter, get_List(r(end_nodes), i),
                                                  get_List(r(end_nodes), j)) : NULL;

        TE_ArcId *arc = arc_id != NULL ? create_ArcId(start_node, arc_id->stimulus_id, end_node) : NULL;
        
        graph->data = *view;
        path = data->engine->reachNode(graph, node, arc);
        
        destroy(node);
        destroy(arc);
        
        if(path != NULL) { goto end_loop; }
    }
    
end_loop:

    graph->data = data;    
    
    destroy(nodes);
    destroy(start_nodes);
    destroy(end_nodes);

    return path;
}

static String* getError_FSM2(TE_Graph *graph)
{
    String *error;
    FSM2_Data *data = (FSM2_Data*)graph->data;

    switch(graph->error)
    {
    case 0:
        {
            return create_String("None");
        }
    default:
        {
            void **view = (void**)get_ArrayList(&data->data, data->i*data->number + data->j);
            graph->data = *view;

            error = concat_String(format_String("[%d,%d] ", data->i, data->j),
                                  data->engine->getError(graph));
            graph->data = data;
            return error;
        }
    }
}

static bool toFile_FSM2(TE_Graph *graph, const char *filename)
{
    int i, j;
    bool status = true;
    FSM2_Data *data = (FSM2_Data*)graph->data;

    for(i = 0;     i < data->number - 1; i++)
    for(j = i + 1; j < data->number;     j++)
    {
        char buffer[strlen(filename)];
        void **view = (void**)get_ArrayList(&data->data, i*data->number + j);
        
        sprintf(buffer, "%s.%d-%d", filename, i, j);
        
        graph->data = *view;
        status &= data->engine->toFile(graph, buffer);
    }
    
    graph->data = data;

    return status;
}

static void finish_FSM2(TE_Graph *graph)
{
    int i, j;
    FSM2_Data *data = (FSM2_Data*)graph->data;

    for(i = 0;     i < data->number - 1; i++)
    for(j = i + 1; j < data->number;     j++)
    {
        void **view = (void**)get_ArrayList(&data->data, i*data->number + j);
        
        graph->data = *view;
        data->engine->finish(graph);
    }

    destroy_ArrayList(&data->data);
    dealloc_Data(data);
    
    graph->data = NULL;
}

static bool generalizes_FSM2(TE_Graph *graph, TE_NodeId *abstract_node_id, TE_NodeId *concrete_node_id)
{
    FSM2_Data *data = (FSM2_Data*)graph->data;

    return abstract_node_id != NULL
        && length_String(r(abstract_node_id->id))
        && charAt_String(r(abstract_node_id->id), 0) == data->delimiter;
}

//------------------------------------------------------------------------------
// FSM2 Engine
//------------------------------------------------------------------------------

TestEngine TE_FSM2 =
{
    /* .name        = */ "netfsm2",
    /* .init        = */ init_FSM2,
    /* .config      = */ config_FSM2,
    /* .start       = */ start_FSM2,
    /* .update      = */ update_FSM2,
    /* .getPath     = */ getPath_FSM2,
    /* .reachNode   = */ reachNode_FSM2,
    /* .getError    = */ getError_FSM2,
    /* .finish      = */ finish_FSM2,
    /* .toFile      = */ toFile_FSM2,
    /* .fromFile    = */ NULL,
    /* .generalizes = */ NULL,
    /* .sysdata     = */ { NULL }
};
