/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <netpfsm.h>
#include <sgl/graph.h>
#include <sgl/utils.h>
#include <utils/assertion.h>

//------------------------------------------------------------------------------
// PFSM Data
//------------------------------------------------------------------------------

typedef struct PFSM_Data
{
    /// The reference to the external graph information.
    TE_GraphInfo *info;
    
    /// The command-line parameters.
    int    argc;
    char **argv;
    
    /// The test length.
    int test_length;
    /// The current length.
    int length;
} PFSM_Data;

static PFSM_Data* alloc_Data(void)
{
    PFSM_Data *data = (PFSM_Data*)malloc(sizeof(PFSM_Data));
    assertion(data != NULL, "alloc_Data: data == NULL");
    
    return data;
}

static void dealloc_Data(PFSM_Data *data)
{
    int i;

    if(data == NULL) { return; }

    if(data->argv != NULL)
    {
        for(i = 0; i < data->argc; i++)
            { free(data->argv[i]); }
            
        free(data->argv);
    }
    
    free(data);
}

static void init_Data(PFSM_Data *data, TE_GraphInfo *info)
{
    data->info = info;

    data->argc = 0;
    data->argv = NULL;
    
    data->test_length = -1;
    data->length = 0;
}

//------------------------------------------------------------------------------
// PFSM Interface
//------------------------------------------------------------------------------

static void processParams_PFSM(TE_Graph *graph, int argc, char **argv)
{
    int i;

    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    PFSM_Data *data2 = (PFSM_Data*)data1->data;

    assertion(argc >= 0,                 "processParams_PFSM: argc < 0");
    assertion(argv != NULL || argc == 0, "processParams_PFSM: argv == NULL && argc > 0");

    if(argc > 0)
        { data2->argv = (char**)malloc(argc * sizeof(char*)); }

    for(i = 0; i < argc; i++)
    {
        if(!strcmp("--cpu", argv[i]))
        {
            assertion(++i < argc, "processParams_PFSM: --cpu should be followed by an integer CPU identifier");
            data2->info->cpu_id = atoi(argv[i]);

            sgl_setId_Graph(data1, data2->info->cpu_id);
        }
        else if(!strcmp("--length", argv[i]))
        {
            assertion(++i < argc, "processParams_PFSM: --length should be followed by an integer number");
            data2->test_length = atoi(argv[i]);
        }
        
        data2->argv[data2->argc++] = strdup(argv[i]);        
    }
}

static void init_PFSM(TE_Graph *graph, int argc, char **argv)
{
    SGL_Graph *data1 = sgl_alloc_Graph();
    PFSM_Data *data2 = alloc_Data();

    sgl_init_Graph(data1, NULL, NULL, (SGL_DataDestructor)dealloc_Data);
    init_Data(data2, &graph->info);

    data1->data = data2;
    graph->data = data1;

    processParams_PFSM(graph, argc, argv);
}

static void config_PFSM(TE_Graph *graph, const char *attr_name, const char *attr_value)
{
    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    PFSM_Data *data2 = (PFSM_Data*)data1->data;
    
    if(!strcmp(TE_ATTR_CPU_ID, attr_name))
    {
        data2->info->cpu_id = atoi(attr_value);
        sgl_setId_Graph(data1, data2->info->cpu_id);
    }
    else
    {
        assertion(false, "config_PFSM: unknown attribute");
    }
}

static TE_UpdateStatus start_PFSM(TE_Graph *graph, TE_NodeId *node_id)
{
    SGL_UpdateStatus status;

    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    PFSM_Data *data2 = (PFSM_Data*)data1->data;

    sgl_addNode_Graph(data1, node_id, &status);

    data2->info->nodes_traversed++;
    data2->info->nodes_traversed_locally++;

    return status == SGL_US_OLD_ALL ? TE_OLD_ALL : TE_NEW_NODE;
}

static TE_UpdateStatus update_PFSM(TE_Graph *graph, TE_UpdateType type,
    TE_ArcId *arc_id, TE_NodeId *new_end_node_id)
{
    SGL_UpdateStatus status;

    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    PFSM_Data *data2 = (PFSM_Data*)data1->data;

    SGL_NodeId *end_node_id = new_end_node_id != NULL ? new_end_node_id : arc_id->end_node_id;
    SGL_Edge *edge = sgl_addEdge_Graph(data1, arc_id->start_node_id, arc_id->stimulus_id, end_node_id, &status);

    return updateInfo_Edge(data2->info, edge, type, status);
}

static TE_ArcList* getPath_PFSM(TE_Graph *graph, TE_NodeId *node_id)
{
    TE_ArcList *path;

    SGL_Node *node;
    SGL_Edge *edge;

    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    PFSM_Data *data2 = (PFSM_Data*)data1->data;
    
    TE_GraphInfo *info = data2->info;

    if(data2->test_length <  0 && info->nodes_finished >= info->nodes_traversed ||
       data2->test_length >= 0 && data2->length >= data2->test_length)
    {
        graph->error = 0;
        return NULL;
    }

    node = sgl_findNode_Graph(data1, node_id);
    assertion(node != NULL, "getPath_PFSM: node == NULL");

    if(sgl_getDegree_Node(node) == 0)
        { return NULL; }

    if(sgl_isFinished_Node(node))
        { edge = sgl_getEdge_Random(data1, node); }
    else
        { edge = sgl_getNewEdge_Random(data1, node); }

    assertion(edge != NULL, "getPath_PFSM: edge == NULL");
    
    path = create_List(&type_TE_ArcId);
    append_List(r(path), create_ArcId(r(node_id), edge->stimulus_id, NULL));

    data2->length++;
    
    return path;
}

static TE_ArcList* reachNode_PFSM(TE_Graph *graph, TE_NodeId *node_id, TE_ArcId *arc_id)
{
    return NULL;
}

static String* getError_PFSM(TE_Graph *graph)
{
    switch(graph->error)
    {
    case 0:
        { return create_String("None"); }
    default:
        { return create_String("Unknown error"); }
    }
}

static bool toFile_PFSM(TE_Graph *graph, const char *filename)
{
    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    PFSM_Data *data2 = (PFSM_Data*)data1->data;
    
    return sgl_toFile_Graph(data1, filename, data2->argc, data2->argv);
}

static void finish_PFSM(TE_Graph *graph)
{
    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    PFSM_Data *data2 = (PFSM_Data*)data1->data;

    sgl_destroy_Graph(data1);
    sgl_dealloc_Graph(data1);

    graph->data = NULL;
}

//------------------------------------------------------------------------------
// PFSM Engine
//------------------------------------------------------------------------------

TestEngine TE_PFSM =
{
    /* .name        = */ "netpfsm",
    /* .init        = */ init_PFSM,
    /* .config      = */ config_PFSM,
    /* .start       = */ start_PFSM,
    /* .update      = */ update_PFSM,
    /* .getPath     = */ getPath_PFSM,
    /* .reachNode   = */ reachNode_PFSM,
    /* .getError    = */ getError_PFSM,
    /* .finish      = */ finish_PFSM,
    /* .toFile      = */ toFile_PFSM,
    /* .fromFile    = */ NULL,
    /* .generalizes = */ NULL,
    /* .sysdata     = */ { NULL }
};
