/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <netrand.h>
#include <atl/object_int.h>
#include <rand/lagged_fibonacci.h>
#include <utils/assertion.h>

#define RAND_RANDOM_INIT(data)       init_MLFG(&data->generator)
#define RAND_RANDOM_SEED(data, seed) seed_MLFG(&data->generator, seed)
#define RAND_RANDOM_NEXT(data)       next_MLFG(&data->generator)

/******************************************************************************/
/**                  RAND Internal Parameters and Macroses                   **/
/******************************************************************************/

/** @typedef RAND_Load
 * The RAND engine generation scheme (for the parallel mode only).
 */
typedef enum RAND_Load
{
    /// Mimimal loading
    RAND_MIN_LOAD = 0x0,
    /// Fixed ratio of the admitted requests is executed.
    RAND_FIX_LOAD = 0x1,
    /// Loading gradually changes from minumum to maximum.
    RAND_VAR_LOAD = 0x2
} RAND_Load;

#define RAND_TEST_LENGTH   1000
#define RAND_PARALLEL_MODE 0
#define RAND_LOADING_MODE  RAND_VAR_LOAD
#define RAND_LOADING_RATIO 50

/******************************************************************************/
/**                             RAND Error Codes                             **/
/******************************************************************************/

/** @typedef RAND_Errno
 * Error codes for the RAND engine.
 */
typedef enum RAND_Errno
{
    /// No errors.
    RAND_ERRNO_NO_ERRORS = 0x0,
    /// Unconnected graph.
    RAND_ERRNO_UNCONNECT = 0x1
} RAND_Errno;

/******************************************************************************/
/**                                RAND Data                                 **/
/******************************************************************************/

/** @typedef RAND_Data
 * Internal representation of the RAND engine state.
 */
typedef struct RAND_Data
{
    /// Graph information (every step is interpreted as a new arc leading to a new node).
    TE_GraphInfo *graph_info;
    /// Test sequence length (number of steps in which a system under test  is called).
    int test_length;
    /// Parallel mode (the first action is a delay, while all others are instant requests).
    int parallel_mode;
    /// Loading mode.
    RAND_Load loading_mode;
    /// Loading ratio.
    int loading_ratio;
    /// Current test length
    int current_length;
    /// Degree of the current node
    int current_degree;
    /// Length of the current path
    int current_path;

    /// Lagged Fibonacci random number generator.
    MLFG generator;
} RAND_Data;

static void init_RAND_Data(RAND_Data *data, TE_GraphInfo *info)
{
    assertion(data != NULL, "init_RAND_Data: data == NULL");
    assertion(info != NULL, "init_RAND_Data: info == NULL");

    data->graph_info = info;

    data->test_length   = RAND_TEST_LENGTH;
    data->parallel_mode = RAND_PARALLEL_MODE;
    data->loading_mode  = RAND_LOADING_MODE;
    data->loading_ratio = RAND_LOADING_RATIO;

    data->current_length = 0;
    data->current_degree = 0;
    data->current_path   = 0;

    RAND_RANDOM_INIT(data);
}

static void processParams_RAND_Data(RAND_Data *data, int argc, char **argv)
{
    int i;

    assertion(data != NULL,              "processParams_RAND_Data: data == NULL");
    assertion(argc >= 0,                 "processParams_RAND_Data: argc < 0");
    assertion(argv != NULL || argc == 0, "processParams_RAND_Data: argv == NULL && argc > 0");

    for(i = 0; i < argc; i++)
    {
        if(!strcmp("--cpu", argv[i]))
        {
            assertion(++i < argc, "processParams_RAND_Data: bad parameter: after --cpu should be an integer CPU identifier");
            data->graph_info->cpu_id = atoi(argv[i]);

            RAND_RANDOM_SEED(data, data->graph_info->cpu_id);
        }
        else if(!strcmp("--length", argv[i]))
        {
            assertion(++i < argc, "processParams_RAND_Data: bad parameter: after --length should be an integer number");
            data->test_length = atoi(argv[i]);
        }
        else if(!strcmp("--parallel", argv[i]))
        {
            data->parallel_mode = 1;
        }
        else if(!strcmp("--sequential", argv[i]))
        {
            data->parallel_mode = 0;
        }
        else if(!strcmp("--min-load", argv[i]))
        {
            data->loading_mode  = RAND_MIN_LOAD;
            data->loading_ratio = 0;
        }
        else if(!strcmp("--max-load", argv[i]))
        {
            data->loading_mode  = RAND_FIX_LOAD;
            data->loading_ratio = 100;
        }
        else if(!strcmp("--fix-load", argv[i]))
        {
            data->loading_mode  = RAND_FIX_LOAD;
            assertion(++i < argc, "processParams_RAND_Data: bad parameter: after --fix-load should be an integer number");
            data->loading_ratio = atoi(argv[i]);
        }
        else if(!strcmp("--var-load", argv[i]))
        {
            data->loading_mode  = RAND_VAR_LOAD;
            data->loading_ratio = 0;
        }
    }
}

static void destroy_RAND_Data(RAND_Data *data)
{
    assertion(data != NULL, "destroy_RAND_Data: data == NULL");
}

/******************************************************************************/
/**                              RAND Interface                              **/
/******************************************************************************/

static void init_RAND(TE_Graph *graph, int argc, char **argv)
{
    RAND_Data *data = (RAND_Data*)malloc(sizeof(RAND_Data));
    assertion(data != NULL, "init_RAND: out of memory");

    init_RAND_Data(data, &graph->info);
    processParams_RAND_Data(data, argc, argv);

    graph->data = data;
}

static void config_RAND(TE_Graph *graph, const char *attr_name, const char *attr_value)
{
    RAND_Data *data = (RAND_Data*)graph->data;

    if(!strcmp(TE_ATTR_CPU_ID, attr_name))
    {
        data->graph_info->cpu_id = atoi(attr_value);
        RAND_RANDOM_SEED(data, data->graph_info->cpu_id);
    }
    else
    {
        assertion(false, "config_RAND: unknown attribute");
    }
}

static TE_UpdateStatus start_RAND(TE_Graph *graph, TE_NodeId *node_id)
{
    RAND_Data *data = (RAND_Data*)graph->data;

    // Increment the node counts
    data->graph_info->nodes_traversed++;
    data->graph_info->nodes_traversed_locally++;

    return TE_NEW_NODE;
}

static TE_UpdateStatus update_RAND(TE_Graph *graph, TE_UpdateType type,
    TE_ArcId *arc_id, TE_NodeId *new_end_node_id)
{
    TE_NodeId *end_node_id;
    RAND_Data *data = graph->data;

    assertion(arc_id != NULL, "update_RAND: arc_id == NULL");

    end_node_id = (new_end_node_id != NULL ? new_end_node_id : arc_id->end_node_id);

    // Add the untraversed arc (while expanding the current node)
    if(end_node_id == NULL)
    {
        // Increment the count of the known arcs
        data->graph_info->arcs_known++;
        // Increment the current node's degree
        data->current_degree++;
    }
    // Add the traversed arc 
    else
    {
        // Increment the arc counts
        data->graph_info->arcs_traversed++;
        data->graph_info->arcs_traversed_locally++;

        // Decrement the current path's length
        data->current_path--;
        assertion(data->current_path >= 0, "update_RAND: too long path");

        // To make the generator expand the node
        if(data->current_path == 0)
            { return TE_NEW_NODE; }
    }

    return TE_OLD_ALL;
}

static void random_reordering(RAND_Data *data, int order[], size_t size)
{
    int i, j, k, t;

    for(i = 0; i < size / 2; i++)
        { t = order[j = RAND_RANDOM_NEXT(data) % size]; order[j] = order[k = RAND_RANDOM_NEXT(data) % size]; order[k] = t; }
}

static TE_ArcList* getPath_RAND(TE_Graph *graph, TE_NodeId *node_id)
{
    int  i, p, num, deg;
    TE_StimulusId stimulus_id;
    TE_ArcList *path = create_List(&type_TE_ArcId);
    RAND_Data  *data = (RAND_Data*)graph->data;
    char sel, arcs[data->current_degree];
    int order[data->current_degree];

    // Traversal has been completed
    if(data->current_length >= data->test_length)
        { graph->error = RAND_ERRNO_NO_ERRORS; return NULL; }

    // Unconnected graph
    if(data->current_degree <= 0)
        { graph->error = RAND_ERRNO_UNCONNECT; return NULL; }

    // Construct a mix of arcs ending with a delay (parallel mode)
    if(data->parallel_mode)
    {
        // Exclude the delay arc
        deg = data->current_degree - 1;

        if(data->loading_mode == RAND_MIN_LOAD)
        {
            data->loading_ratio = 0;
        }
        else if(data->loading_mode == RAND_VAR_LOAD)
        {
            if(data->current_length < data->test_length / 2)
                { data->loading_ratio = (data->current_length * 200) / data->test_length; }
            else
                { data->loading_ratio = 100; }
        }

        if(data->current_degree == 1)
            { data->loading_ratio = 100; }
        else if(data->loading_ratio < 100 / deg)
            { data->loading_ratio = 100 / deg; }

        // Maximal loading
        if(data->loading_ratio == 100)
        {
            // Choose all possible arcs
            memset(&arcs[0], 1, sizeof(arcs));
        }
        // Ratio-based loading
        else
        {
            // Include or exclude: that is the question
            sel = data->loading_ratio <= 50 ? 1 : 0;

            // Number of arcs to choose (excluding a delay at the end)
            if((num = (p = data->loading_ratio * deg) / 100) < deg)
                { num += ((RAND_RANDOM_NEXT(data) % 100) < (p % 100) ? 1 : 0); }

            // The delay arc is always selected
            arcs[0] = 1; memset(&arcs[1], !sel, sizeof(arcs) - 1);
            if(!sel) { num = deg - num; }

            // Choose the rest of the arcs randomly
            while(num)
            {
                if(arcs[i = (RAND_RANDOM_NEXT(data) % deg) + 1] != sel)
                    { arcs[i] = sel; num--; }
            }
        }

        // Compose the arcs choosen
        for(i = num = 0; i < data->current_degree; i++)
        {
            if(arcs[p = (i + 1) % data->current_degree])
                { order[num++] = p; }
        }

        // Reorder the arcs leaving the delay arc at the end
        random_reordering(data, order, num - 1);

        // Several requests (index > 0) + a delay (index = 0) at the end
        for(i = 0; i < num; i++)
        {
            stimulus_id.index = order[i];

            if(arcs[stimulus_id.index])
                { append_List(r(path), create_ArcId(r(node_id), stimulus_id, NULL)); }
        }
    }
    // Construct a random single-arc path (sequential mode)
    else
    {
        stimulus_id.index = RAND_RANDOM_NEXT(data) % data->current_degree;
        append_List(r(path), create_ArcId(r(node_id), stimulus_id, NULL));
    }

    data->current_length++;
    data->current_degree = 0;
    data->current_path = size_List(r(path));

    return path;
}

static TE_ArcList* reachNode_RAND(TE_Graph *graph, TE_NodeId *node_id, TE_ArcId *arc_id)
{
    graph->error = RAND_ERRNO_UNCONNECT;
    return NULL;
}

static String* getError_RAND(TE_Graph *graph)
{
    switch(graph->error)
    {
    case RAND_ERRNO_NO_ERRORS:
        { return create_String("None"); }
    case RAND_ERRNO_UNCONNECT:
        { return create_String("Unconnected graph"); }
    default:
        { return create_String("Unknown error"); }
    }
}

static void finish_RAND(TE_Graph *graph)
{
    RAND_Data *data = (RAND_Data*)graph->data;

    if(data != NULL)
    {
        destroy_RAND_Data(data);
        free(data);
        graph->data = NULL;
    }
}

/******************************************************************************/
/**                               RAND Engine                                **/
/******************************************************************************/

TestEngine TE_RAND =
{
    /* .name        = */ "netrand",
    /* .init        = */ init_RAND,
    /* .config      = */ config_RAND,
    /* .start       = */ start_RAND,
    /* .update      = */ update_RAND,
    /* .getPath     = */ getPath_RAND,
    /* .reachNode   = */ reachNode_RAND,
    /* .getError    = */ getError_RAND,
    /* .finish      = */ finish_RAND,
    /* .toFile      = */ NULL,
    /* .fromFile    = */ NULL,
    /* .generalizes = */ NULL,
    /* .sysdata     = */ { NULL }
};

