/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __SGL_GRAPH_H_INCLUDED__
#define __SGL_GRAPH_H_INCLUDED__

#include <atl/arraylist.h>
#include <atl/hashmap.h>
#include <atl/string.h>
#include <netfsm/generator.h>
#include <rand/lagged_fibonacci.h>

//------------------------------------------------------------------------------
// Graph Data Types
//------------------------------------------------------------------------------

typedef struct TE_NodeId SGL_NodeId;

typedef struct TE_StimulusId SGL_StimulusId;

typedef struct SGL_Node SGL_Node;

typedef struct SGL_Edge SGL_Edge;

typedef struct SGL_Graph SGL_Graph;

typedef int (*SGL_NodeChecker)(SGL_Node *current_node, SGL_Node *target_node);

typedef int (*SGL_EdgeChecker)(SGL_Edge *edge);

typedef void (*SGL_DataDestructor)(void *data);

typedef enum SGL_UpdateStatus
{
    SGL_US_OLD_ALL        = 0x00,
    SGL_US_NEW_START_NODE = 0x01,
    SGL_US_NEW_END_NODE   = 0x02,
    SGL_US_NEW_NODE       = SGL_US_NEW_START_NODE | SGL_US_NEW_END_NODE,
    SGL_US_FIN_START_NODE = 0x04,
    SGL_US_NEW_EDGE       = 0x08,
    SGL_US_NEW_LOOP       = 0x10
} SGL_UpdateStatus;

struct SGL_Node
{
    /// The internal node identifier.
    int id;
    /// The external node identifier.
    SGL_NodeId *node_id;
    /// The outgoing edges.
    ArrayList edges;

    /// The weight of the node (for search optimization).
    int weight;
    /// The distance from the current node (for search optimization)
    int distance;    

    /// The number of new edges.
    int new_edge_count;
    /// The index of the recent new edge.
    int new_edge_index;
    
    /// The reference to the graph.
    SGL_Graph *graph;
    /// The associated data.
    void *data;
};

struct SGL_Edge
{
    /// The stimulus identifier (label).
    SGL_StimulusId stimulus_id;
    /// The start node.
    SGL_Node *start_node;
    /// The end nodes.
    ArrayList end_nodes;
    /// The number of loop edges.
    int add_loop_count;
    /// The number of usual edges.
    int add_edge_count;

    /// The associated data.
    void *data;
};

struct SGL_Graph
{
    /// The graph identifier (it is used in distributed traversal).
    int id;

    /// The list of nodes.
    ArrayList nodes;
    /// Maps name to id.
    HashMap *hash_map;
    /// The result of the recent search (a path in reverse order).
    ArrayList path;

    /// The lagged Fibonacci random number generator.
    MLFG generator;
    
    /// The node data destructor.
    SGL_DataDestructor node_destructor;
    /// The edge data destructor.
    SGL_DataDestructor edge_destructor;
    /// The graph data destructor.
    SGL_DataDestructor graph_destructor;
    
    /// The associated data.
    void* data;
};

//------------------------------------------------------------------------------
// Graph Node Macros & Functions
//------------------------------------------------------------------------------
#define SGL_MINIMAL_WEIGHT     0x00000000
#define SGL_INITIAL_WEIGHT     0x00000000
#define SGL_MAXIMAL_WEIGHT     0x3fffFFFF

#define sgl_add_Weight(x, y) \
    (((x) + (y) >= SGL_MAXIMAL_WEIGHT) ? SGL_MAXIMAL_WEIGHT : (x) + (y))

#define sgl_sub_Weight(x, y) \
    (((x) >= (y) + SGL_MINIMAL_WEIGHT) ? (x) - (y) : SGL_MINIMAL_WEIGHT)

#define sgl_getId_Node(node) \
    ((node)->id)

#define sgl_getNodeId_Node(node) \
    ((node) != NULL ? (node)->node_id : NULL)

#define sgl_getName_Node(node) \
    (toCharArray_String(r((node)->node_id->id)))

#define sgl_getDegree_Node(node) \
    ((node)->edges.size)

#define sgl_getEdge_Node(node, i) \
    ((SGL_Edge*)get_ArrayList(&(node)->edges, i))

#define sgl_isFinished_Node(node) \
    ((node)->new_edge_count == 0)

#define sgl_getNewEdgeCount_Node(node) \
    ((node)->new_edge_count)

#define sgl_getWeight_Node(node) \
    ((node)->weight)

#define sgl_setWeight_Node(node, w) \
    ((node)->weight = w)
    
SGL_Node* sgl_alloc_Node(void);

void sgl_dealloc_Node(SGL_Node* node);

void sgl_init_Node(SGL_Node *node, int id, SGL_NodeId *node_id, SGL_Graph *graph);

void sgl_initWithData_Node(SGL_Node *node, int id, SGL_NodeId *node_id, SGL_Graph *graph, void *data);

void sgl_destroy_Node(SGL_Node *node);

SGL_Edge* sgl_findEdge_Node(SGL_Node *node, SGL_StimulusId stimulus_id);

SGL_Edge* sgl_addEdge_Node(SGL_Node *start_node, SGL_StimulusId stimulus_id,
    SGL_Node *end_node, SGL_UpdateStatus *status);
    
SGL_Edge* sgl_findNewEdge_Node(SGL_Node *node, int start_search_index);

void sgl_updateWeight_Node(SGL_Node *node, SGL_NodeChecker node_checker, SGL_EdgeChecker edge_checker);

//------------------------------------------------------------------------------
// Graph Edge Macros & Functions
//------------------------------------------------------------------------------

#define sgl_isValid_Edge(edge) \
    ((edge)->start_node != NULL)

#define sgl_getStimulusId_Edge(edge) \
    ((edge)->stimulus_id)

#define sgl_getLabel_Edge(edge) \
    ((edge)->stimulus_id.index)

#define sgl_getStartNode_Edge(edge) \
    ((edge)->start_node)

#define sgl_getEndNode_Edge(edge, i) \
    (*((SGL_Node**)get_ArrayList(&(edge)->end_nodes, (i))))

#define sgl_getDegree_Edge(edge) \
    ((edge)->end_nodes.size)

#define sgl_isNew_Edge(edge) \
    ((edge)->end_nodes.size == 0 && (edge)->add_loop_count == 0)

#define sgl_isDeterministic_Edge(edge) \
    ((edge)->end_nodes.size == 1 && (edge)->add_loop_count == 0 || ((edge)->end_nodes.size == 0 && (edge)->add_loop_count > 0))

SGL_Edge* sgl_alloc_Edge(void);

void sgl_dealloc_Edge(SGL_Edge* edge);

void sgl_init_Edge(SGL_Edge *edge, SGL_Node *start_node, SGL_StimulusId stimulus_id, SGL_Node *end_node);

void sgl_initWithData_Edge(SGL_Edge *edge, SGL_Node *start_node,
    SGL_StimulusId stimulus_id, SGL_Node *end_node, void *data);

void sgl_destroy_Edge(SGL_Edge *edge);

//------------------------------------------------------------------------------
// Graph Macros & Functions
//------------------------------------------------------------------------------

#define sgl_getId_Graph(graph) \
    ((graph)->id)

#define sgl_setId_Graph(graph, i) \
    sgl_seed_Random((graph), ((graph)->id = (i)))
        
#define sgl_getSize_Graph(graph) \
    ((graph)->nodes.size)

#define sgl_getNode_Graph(graph, i) \
    ((SGL_Node*)get_ArrayList(&(graph)->nodes, (i)))

SGL_Graph* sgl_alloc_Graph(void);

void sgl_dealloc_Graph(SGL_Graph* graph);

void sgl_init_Graph
(
    SGL_Graph *graph,
    SGL_DataDestructor node_destructor,
    SGL_DataDestructor edge_destructor,
    SGL_DataDestructor graph_destructor
);

void sgl_initWithData_Graph
(
    SGL_Graph *graph, void *data,
    SGL_DataDestructor node_destructor,
    SGL_DataDestructor edge_destructor,
    SGL_DataDestructor graph_destructor
);

void sgl_destroy_Graph(SGL_Graph *graph);

SGL_Node* sgl_findNode_Graph(SGL_Graph *graph, SGL_NodeId *node_id);

SGL_Node* sgl_addNode_Graph(SGL_Graph *graph, SGL_NodeId *node_id, SGL_UpdateStatus *status);

SGL_Edge* sgl_addEdge_Graph(SGL_Graph *graph, SGL_NodeId *start_node_id,
    SGL_StimulusId stimulus_id, SGL_NodeId *end_node_id, SGL_UpdateStatus *status);
    
SGL_Node* sgl_findNonFinishedNode_Graph(SGL_Graph *graph);

//------------------------------------------------------------------------------
// Path Finder Macros & Functions
//------------------------------------------------------------------------------

int sgl_targetNodeChecker(SGL_Node *current_node, SGL_Node *target_node);

int sgl_nonFinishedNodeChecker(SGL_Node *current_node, SGL_Node *target_node);

int sgl_deterministicEdgeChecker(SGL_Edge *edge);

int sgl_emptyEdgeChecker(SGL_Edge *edge);

int sgl_findPath_TargetNode_Graph
(
    SGL_Graph      *graph,
    SGL_Node       *source_node,
    SGL_Node       *target_node,
    SGL_EdgeChecker edge_checker,
    int             randomize
);

int sgl_findPath_NonFinishedNode_Graph
(
    SGL_Graph      *graph,
    SGL_Node       *source_node,
    SGL_EdgeChecker edge_checker,
    int             randomize
);

int sgl_findPath_NodeChecker_Graph
(
    SGL_Graph      *graph,
    SGL_Node       *source_node,
    SGL_NodeChecker node_checker,
    SGL_EdgeChecker edge_checker,
    int             randomize
);

//------------------------------------------------------------------------------
// Randomization Macros & Functions
//------------------------------------------------------------------------------

#define sgl_init_Random(graph) \
    init_MLFG(&(graph)->generator)

#define sgl_seed_Random(graph, seed) \
    seed_MLFG(&(graph)->generator, (seed))

#define sgl_next_Random(graph) \
    next_MLFG(&(graph)->generator)

SGL_Edge* sgl_getNewEdge_Random(SGL_Graph *graph, SGL_Node *node);

SGL_Edge* sgl_getEdge_Random(SGL_Graph *graph, SGL_Node *node);

//------------------------------------------------------------------------------
// Graph Printer Macros & Functions
//------------------------------------------------------------------------------

int sgl_toFile_Graph(SGL_Graph *graph, const char *filename, int argc, char **argv);

#endif
