/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <utils/compressor.h>

#include <memory.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>

#include <utils/assertion.h>

#undef  ENCODE_4BITS
#define ENCODE_6BITS

#define BUFFER_SIZE 512

/******************************************************************************/
/**                        Node Compression Functions                        **/
/******************************************************************************/

static char encode_4bits_table[] =
{
    '0', '1', '2', '3',
    '4', '5', '6', '7',
    '8', '9', 'a', 'b',
    'c', 'd', 'e', 'f'
};

static char encode_6bits_table[] =
{
    '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', 'a', 'b', 'c', 'd', 'e', 'f',
    'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
    'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
    'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D',
    'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',
    'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
    'U', 'V', 'W', 'X', 'Y', 'Z', '_', '-'
};

static char* encode_4bits(const char *src, size_t src_size, char dst[], size_t dst_size)
{
	int bytes_required = 2 * src_size + 1 /* zero byte */;
	assertion(bytes_required <= dst_size, "encode_4bits: destination buffer is too small");

	dst[dst_size = 2 * src_size] = 0;
	for(--src_size; !(src_size & 0x10000000); src_size--)
	{
		unsigned char src_byte = src[src_size];

		dst[--dst_size] = encode_4bits_table[src_byte >>  4];
		dst[--dst_size] = encode_4bits_table[src_byte & 0xf];
	}

	return dst;
}

static char* encode_6bits(const char *src, size_t src_size, char dst[], size_t dst_size)
{
	int i = 0;
	int j = 0;
	int k = 3 * (src_size / 3);

	unsigned char src_byte_0, src_byte_1, src_byte_2;

	int bytes_required = (4 * src_size) / 3 + 3 /* if not multiple of 3 */ + 1 /* zero byte */;
	assertion(bytes_required <= dst_size, "encode_6bits: destination buffer is too small");

	while(i < k)
	{
		src_byte_0 = src[i++];
		src_byte_1 = src[i++];
		src_byte_2 = src[i++];

		dst[j++] = encode_6bits_table[(src_byte_0) & 0x3f];
		dst[j++] = encode_6bits_table[(src_byte_0 >> 6 | src_byte_1 << 2) & 0x3f];
		dst[j++] = encode_6bits_table[(src_byte_1 >> 4 | src_byte_2 << 4) & 0x3f];
		dst[j++] = encode_6bits_table[(src_byte_2 >> 2) & 0x3f];
	}

	if(i < src_size)
	{
		src_byte_0 = src[i++];
		dst[j++] = encode_6bits_table[(src_byte_0) & 0x3f];

		if(i < src_size)
		{
			src_byte_1 = src[i++];
			dst[j++] = encode_6bits_table[(src_byte_0 >> 6 | src_byte_1 << 2) & 0x3f];
			dst[j++] = encode_6bits_table[(src_byte_1 >> 4) & 0x3f];
		}
		else
		{
			dst[j++] = encode_6bits_table[(src_byte_0 >> 6) & 0x3f];
		}
	}

	dst[j] = 0;

	return dst;
}

static char* encode(const char *src, size_t src_size, char dst[], size_t dst_size)
{
#if defined(ENCODE_4BITS)
	return encode_4bits(src, src_size, dst, dst_size);
#elif defined(ENCODE_6BITS)
	return encode_6bits(src, src_size, dst, dst_size);
#else
	#error "unknown encoding"
#endif
}

/******************************************************************************/
/**                      XOR-Based Compression Functions                     **/
/******************************************************************************/

static String* compress_XOR(Object *node, size_t size)
{
	int i, length;
	String *string = toString(node);

	char src_buffer[size];
	char dst_buffer[BUFFER_SIZE];

	const char *array = toCharArray_String(r(string));

	if((length = length_String(r(string))) < size)
	    { return string; }

	memset(&src_buffer[0], 0, size);

	for(i = 0; i < length; i++)
		{ src_buffer[i % size] ^= array[i]; }

	destroy(string);

	return create_String(encode(src_buffer, size, dst_buffer, sizeof(dst_buffer)));
}

String* compress_XOR16(Object *node) { return compress_XOR(node, 2); }
String* compress_XOR32(Object *node) { return compress_XOR(node, 4); }
String* compress_XOR64(Object *node) { return compress_XOR(node, 8); }

/******************************************************************************/
/**                     CRC-Based Compression Functions                      **/
/******************************************************************************/

#define DEFINE_CRC_FUNCTION(width, poly, init) \
    static uint##width##_t crc##width##_table[256]; \
    static int crc##width##_table_initialized = 0; \
    \
    static void initialize_crc##width##_table(void) \
    { \
        int i, j; \
        uint##width##_t crc; \
        for(i = 0; i < 256; i++) \
        { \
            crc = i; \
            for(j = 0; j < 8; j++) { crc = crc & 1 ? (crc >> 1) ^ poly : crc >> 1; } \
            crc##width##_table[i] = crc; \
        } \
        crc##width##_table_initialized = 1; \
    } \
    \
    static uint##width##_t crc##width(const char *src, size_t src_size) \
    { \
        uint##width##_t crc = init; \
        if(!crc##width##_table_initialized) { initialize_crc##width##_table(); } \
        while(src_size--) { crc = (crc >> 8) ^ crc##width##_table[(crc ^ *src++) & 0xff]; } \
        return crc; \
    }

/*
 * Name  : CRC-16
 * Poly  : 0x8005 (x^16+x^15+x^2+1) (Mirror: 0xa001)
 * Init  : 0xFFFF
 * XorOut: 0x0000
 */
DEFINE_CRC_FUNCTION(16, 0xa001u, 0xffffu)

/*
 * Name   : CRC-32
 * Poly   : 0x04C11DB7 (x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x+1) (Mirror: 0xEDB88320)
 * Init   : 0xFFFFFFFF
 * XorOut : 0x00000000
 */
DEFINE_CRC_FUNCTION(32, 0xedb88320u, 0xffffFFFFu)

/*
 * Name   : CRC-64
 * Poly   : 0x000000000000001B (x^64+x^4+x^3+x+1) (Mirror: 0xD800000000000000)
 * Init   : 0xFFFFFFFFFFFFFFFF
 * XorOut : 0x0000000000000000
 */
DEFINE_CRC_FUNCTION(64, 0xd800000000000000ull, 0xffffFFFFffffFFFFull)

#define DEFINE_COMPRESS_CRC(width) \
    String* compress_CRC##width(Object *node) \
    { \
        String *string = toString(node); /* node is destroyed */ \
		uint##width##_t crc = crc##width(toCharArray_String(r(string)), length_String(r(string))); \
        char dst_buffer[BUFFER_SIZE]; \
		destroy(string); \
		return create_String(encode((const char*)&crc, sizeof(crc), dst_buffer, sizeof(dst_buffer))); \
    }

DEFINE_COMPRESS_CRC(16)
DEFINE_COMPRESS_CRC(32)
DEFINE_COMPRESS_CRC(64)
